import os
import shutil
import subprocess
import tempfile
from pathlib import Path

import pandas as pd
import yaml

from fiddler.core_objects import (
    DatasetInfo,
    DeploymentOptions,
    DeploymentType,
    ModelInfo,
)
from fiddler.fiddler_api import FiddlerApi
import logging

fiddler_core_version = '0.1.1'
LOG = logging.getLogger(__name__)


def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


def dockerfile(args):
    return f'''
    # syntax=docker/dockerfile:1
    FROM fiddlerai/fiddler-core:{fiddler_core_version}

    RUN mkdir -p /app/{args.org}/{args.project}/{args.model}

    COPY . /app/{args.org}/{args.project}/{args.model}

    RUN pip install -r /app/{args.org}/{args.project}/{args.model}/requirements.txt

    ENV FAR_MODEL_PATH {args.org}/{args.project}/{args.model}
    ENV SERVICE_RUN_MODE local

    CMD ["/bin/bash", "/app/runit.sh"]
    '''


def call_cmd(cmd, cwd):
    process = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
        cwd=cwd,
    )

    return_code = None

    while True:
        output = process.stdout.readline()
        LOG.info(output.strip())
        # Do something else
        return_code = process.poll()
        if return_code is not None:
            LOG.info(f'RETURN CODE: {return_code}')
            # Process has finished, read rest of the output
            for output in process.stdout.readlines():
                LOG.info(output.strip())
            for output in process.stderr.readlines():
                LOG.info(output.strip())
            break

    if return_code == 0:
        LOG.info('cmd successful')
    else:
        raise ValueError('cmd failed')


def build_cmd(args):
    LOG.info(f'building archive for  {args.org} {args.project} {args.model}')

    source = Path(args.source)
    if not source.is_dir():
        raise ValueError('source is not a directory')

    if not Path(source, 'requirements.txt').is_file():
        raise ValueError(f'requirements.txt not found in {source}')

    if args.train:
        if not Path(source, 'train.py').is_file():
            raise ValueError(f'train.py not found in {source}')
    else:
        if not Path(source, 'package.py').is_file():
            raise ValueError(f'package.py not found in {source}')
        if not Path(source, 'model.yaml').is_file():
            raise ValueError(f'model.yaml not found in {source}')
        if not Path(source, 'dataset.csv').is_file():
            raise ValueError(f'dataset.csv not found in {source}')
        if not Path(source, 'dataset.yaml').is_file():
            raise ValueError(f'dataset.yaml not found in {source}')

    with tempfile.TemporaryDirectory() as tmp:
        copytree(source, Path(tmp))
        docker_template = dockerfile(args)
        with open(tmp / Path('Dockerfile'), 'w') as output:
            output.write(docker_template)

        LOG.info(os.listdir(tmp))
        name = f'{args.org}-{args.project}-{args.model}'

        cmd = ['docker', 'build', '--file=Dockerfile', f'--tag={name}', '.']
        call_cmd(cmd, tmp)
        return 'done'


def run_cmd(args):
    name = f'{args.org}-{args.project}-{args.model}'

    cmd = [
        'docker',
        'run',
        '-d',
        '--rm',
        '--publish=5100:5100',
        f'--name={name}',
        f'{name}:latest',
    ]

    call_cmd(cmd, '.')


def register_cmd(args):
    client = FiddlerApi(args.endpoint, args.org, args.auth_key)
    if args.model in client.list_models(args.project):
        raise ValueError('model already exists in project')

    source = Path(args.source)
    if not source.is_dir():
        raise ValueError('source is not a directory')

    if not Path(source, 'model.yaml').is_file():
        raise ValueError(f'model.yaml not found in {source}')

    with open(Path(source, 'model.yaml')) as f:
        model_info = ModelInfo.from_dict(yaml.safe_load(f))

    with open(Path(source, 'dataset.yaml')) as f:
        dataset_info = DatasetInfo.from_dict(yaml.safe_load(f))

    if args.dataset:
        dataset_id = args.dataset
    else:
        dataset_id = f'{args.model}_dataset'

    try:
        dataset_info = client.get_dataset_info(args.project, dataset_id)
        LOG.info('Using existing dataset')
    except Exception as e:
        LOG.info(f'Dataset not found, uploading dataset {dataset_id}')
        with open(Path(source, 'dataset.csv')) as f:
            dataset_df = pd.read_csv(f)
            ds_dict = {'dataset': dataset_df}
            result = client.upload_dataset(
                project_id=args.project,
                dataset=ds_dict,
                dataset_id=dataset_id,
                info=dataset_info,
            )
            LOG.info(result)

    deployment_options = DeploymentOptions(
        deployment_type=DeploymentType.FAR,
        image=args.image,
        cpus=0.5,
        memory='512m',
    )

    result = client.register_model(
        args.project, args.model, dataset_id, model_info, deployment_options
    )

    LOG.info(result)



import configparser
import contextlib
import json
import os
import textwrap
import time
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import requests
from packaging import version
from requests.exceptions import Timeout
from requests_toolbelt.multipart.encoder import MultipartEncoder

from fiddler.utils import logging

try:
    from simplejson import JSONDecodeError
except ImportError:
    from json import JSONDecodeError

from ._version import __version__
from .utils.exceptions import JSONException, ResourceNotFound

LOG = logging.getLogger(__name__)

SIMPLE_RES_API = ['slice_query', 'parse_slice_query']


class Connection:

    SUCCESS_STATUS = 'SUCCESS'
    FAILURE_STATUS = 'FAILURE'
    AUTH_HEADER_KEY = 'Authorization'
    STREAMING_HEADER_KEY = 'X-Fiddler-Results-Format'
    FIDDLER_ARGS_KEY = '__fiddler_args__'
    ROUTING_HEADER_KEY = 'x-fdlr-fwd'

    """Broker of all connections to the Fiddler API.
    Conventions:
        - Exceptions are raised for FAILURE reponses from the backend.
        - Methods beginning with `list` fetch lists of ids (e.g. all model ids
            for a project) and do not alter any data or state on the backend.
        - Methods beginning with `get` return a more complex object but also
            do not alter data or state on the backend.
        - Methods beginning with `run` invoke model-related execution and
            return the result of that computation. These do not alter state,
            but they can put a heavy load on the computational resources of
            the Fiddler engine, so they should be paralellized with care.
        - Methods beginning with `delete` permanently, irrevocably, and
            globally destroy objects in the backend. Think "rm -rf"
        - Methods beginning with `upload` convert and transmit supported local
            objects to Fiddler-friendly formats loaded into the Fiddler engine.
            Attempting to upload an object with an identifier that is already
            in use will result in an exception being raised, rather than the
            existing object being overwritten. To update an object in the
            Fiddler engine, please call both the `delete` and `upload` methods
            pertaining to the object in question.

    :param url: The base URL of the API to connect to. Usually either
        https://dev.fiddler.ai (cloud) or http://localhost:4100 (onebox)
    :param org_id: The name of your organization in the Fiddler platform
    :param auth_token: Token used to authenticate. Your token can be
        created, found, and changed at <FIDDLER URL>/settings/credentials.
    :param proxies: optionally, a dict of proxy URLs. e.g.,
                    proxies = {'http' : 'http://proxy.example.com:1234',
                               'https': 'https://proxy.example.com:5678'}
    :param verbose: if True, api calls will be logged verbosely,
                    *warning: all information required for debugging will be
                    logged including the auth_token.
    """

    def __init__(
        self,
        url=None,
        org_id=None,
        auth_token=None,
        proxies=None,
        verbose=False,
        timeout: int = None,
        verify=False,
    ):
        self._verify = verify
        if Path('fiddler.ini').is_file():
            config = configparser.ConfigParser()
            config.read('fiddler.ini')
            info = config['FIDDLER']
            if not url:
                url = info['url']
            if not org_id:
                org_id = info['org_id']
            if not auth_token:
                auth_token = info['auth_token']

        url = url.rstrip('/')  # we want url without trailing '/'

        # use session to preserve session data
        self.session = requests.Session()
        self.session.verify = self._verify
        if proxies:
            assert isinstance(proxies, dict)
            self.session.proxies = proxies
        self.adapter = requests.adapters.HTTPAdapter(
            pool_connections=25,
            pool_maxsize=25,
        )
        self.session.mount(url, self.adapter)
        self.url = url
        self.org_id = org_id
        self.auth_header = {self.AUTH_HEADER_KEY: f'Bearer {auth_token}'}
        self.streaming_header = {self.STREAMING_HEADER_KEY: 'application/jsonlines'}
        self.verbose = verbose
        self.strict_mode = True
        self.capture_server_log = False
        self.last_server_log = None
        self.timeout = timeout
        self.check_connection()

    def check_connection(self, check_client_version=True, check_server_version=True):
        try:
            path = ['get_supported_features', self.org_id]
            self.call(path, is_get_request=True)
            return 'OK'
        except requests.exceptions.ConnectionError as e:
            LOG.exception(
                'CONNECTION CHECK FAILED: Unable to connect with '
                'to Fiddler. Are you sure you have the right URL?'
            )
            raise e
        except Exception as e:
            LOG.exception(
                f'API CHECK FAILED: Able to connect to Fiddler, '
                f'but request failed with message:\n"{str(e)}"'
            )
            raise e

    @staticmethod
    def _get_routing_header(path_base: str) -> Dict[str, str]:
        """Returns the proper header so that a request is routed to the correct
        service."""
        # @todo: Given we check for the service in admin service using fiddler-ingress yaml, we should not be using this here. Just remove it ensuring no side effects.
        executor_service_bases = (
            'dataset_predictions',
            'execute',
            'executor',
            'explain',
            'explain_by_row_id',
            'fairness',
            'feature_importance',
            'generate',
            'new_project',
            'trigger_pre_computation',
            'precache_globals',
        )
        if path_base in executor_service_bases:
            return {Connection.ROUTING_HEADER_KEY: 'executor_service'}
        else:
            return {Connection.ROUTING_HEADER_KEY: 'data_service'}

    def _form_request(
        self,
        path: List[str],
        is_get_request: bool = None,
        json_payload: Any = None,
        stream: bool = False,
        files: Optional[List[Path]] = None,
        context_stack=None,
        endpoint=None,
    ):
        if is_get_request:
            req = requests.Request('GET', endpoint)
        else:
            # if uploading files, we use a multipart/form-data request and
            # dump the json_payload to be the special "fiddler args"
            # as a json object in the form

            if files is not None:
                # open all the files into the context manager stack
                opened_files = {
                    fpath.name: context_stack.enter_context(fpath.open('rb'))
                    for fpath in files
                }
                #
                # NOTE: there are a lot LOT of ways to do this wrong with
                # `requests`
                #
                # Take a look here (and at the thread linked) for more info:
                # https://stackoverflow.com/a/35946962
                #
                # And here: https://stackoverflow.com/a/12385661
                #
                form_data: Dict[str, Tuple[Optional[str], str, str]] = {
                    **{
                        self.FIDDLER_ARGS_KEY: (
                            None,  # filename
                            json.dumps(json_payload),  # data
                            'application/json',  # content_type
                        )
                    },
                    **{
                        fpath.name: (
                            fpath.name,  # filename
                            opened_files[fpath.name],  # data
                            'application/octet-stream',  # content_type
                        )
                        for fpath in files
                    },
                }
                # Take a look at this for how to use MultipartEncoder
                # https://stackoverflow.com/questions/12385179/how-to-send-a-multipart-form-data-with-requests-in-python/12385661#12385661

                mp_encoder = MultipartEncoder(form_data)
                req = requests.Request(
                    'POST',
                    endpoint,
                    data=mp_encoder,
                    headers={'Content-Type': mp_encoder.content_type},
                )
            else:
                req = requests.Request(
                    'POST',
                    endpoint,
                    data=json.dumps(json_payload, allow_nan=True),
                    headers={'Content-Type': 'application/json'},
                )

        # add necessary headers
        # using prepare_request from session to keep session data
        req = self.session.prepare_request(req)

        added_headers = dict()
        added_headers.update(self.auth_header)
        added_headers.update(self._get_routing_header(path[0]))
        if self.capture_server_log:
            added_headers['X-Fiddler-Logs-Level'] = 'DEBUG'
        if stream:
            added_headers.update(self.streaming_header)
        req.headers = {**added_headers, **req.headers}

        return req

    @staticmethod
    def _handle_fail_res(res, endpoint, api_name):
        """
        Raises an actionable error message for a response with a status code > 200
        """
        try:
            # catch auth failure
            json_response = res.json()
            message = json_response.get('message')
            if endpoint == '':
                error_msg = message
            elif api_name in SIMPLE_RES_API:
                error_msg = message
            else:
                error_msg = (
                    f'API call to {endpoint} failed with status {res.status_code}:'
                    f' The full response message was {message}'
                )
            # catch resource not found failure
            if res.status_code == 404:
                error = ResourceNotFound(message=message)
            elif res.status_code == 401:
                # More specific error message
                error_msg = (
                    'API call failed with status 401: '
                    'Authorization Required. '
                    'Do you have the right `org_id` and `auth_token`?'
                )
                error = RuntimeError(error_msg)
            else:
                error = JSONException(
                    status=json_response.get('status'),
                    message=error_msg,
                    stacktrace=json_response.get('stacktrace'),
                    logs=json_response.get('logs'),
                )
        except JSONDecodeError:
            if api_name in SIMPLE_RES_API:
                error_msg = res.txt
            else:
                error_msg = (
                    f'API call to {endpoint} failed with status {res.status_code}:'
                    f' The full response message was {res.text}'
                )
            error = RuntimeError(error_msg)
        LOG.debug(error_msg)
        raise error

    def call(
        self,
        path: List[str],
        json_payload: Any = None,
        files: Optional[List[Path]] = None,
        is_get_request: bool = False,
        stream: bool = False,
        timeout: Optional[int] = None,
        num_tries: int = 1,
    ):
        """Issues a request to the API and returns the result,
        logigng and handling errors appropriately.

        Raises a RuntimeError if the response is a failure or cannot be parsed.
        Does not handle any ConnectionError exceptions thrown by the `requests`
        library.

        Note: Parameters `timeout` and `num_tries` are currently only utilized in a workaround
        for a bug involving Mac+Docker communication. See: https://github.com/docker/for-mac/issues/3448
        """
        timeout = timeout if timeout else self.timeout
        res: Optional[requests.Response] = None
        assert self.url is not None, 'self.url unexpectedly None'
        endpoint = '/'.join([self.url] + path)

        # set up a context manager to open files
        with contextlib.ExitStack() as context_stack:
            request_type = 'GET' if is_get_request else 'POST'

            request_excerpt: Optional[str] = None
            if json_payload:
                request_excerpt = textwrap.indent(
                    json.dumps(json_payload, indent=2)[:2048], '  '
                )

            if self.verbose:
                LOG.info(
                    f'running api call as {request_type} request\n'
                    f'to {endpoint}\n'
                    f'with headers {self.auth_header}\n'
                    f'with payload {request_excerpt}'
                )

            req = self._form_request(
                path=path,
                is_get_request=is_get_request,
                json_payload=json_payload,
                stream=stream,
                files=files,
                context_stack=context_stack,
                endpoint=endpoint,
            )

            # log the raw request
            raw_request_info = (
                f'Request:\n'
                f'  url: {req.url}\n'
                f'  method: {req.method}\n'
                f'  headers: {req.headers}'
            )
            LOG.debug(raw_request_info)

            if os.environ.get('REQUESTS_CA_BUNDLE'):
                self.session.verify = os.environ.get('REQUESTS_CA_BUNDLE')

            # send the request using session to carry auth info from login
            if 'FIDDLER_RETRY_PUBLISH' in os.environ and str(
                os.environ['FIDDLER_RETRY_PUBLISH']
            ).lower() in ['yes', 'y', 'true', '1']:
                # Experimental retry path needed in case of Mac-Docker communication bug.
                # Likely only needed in case of Onebox Mac-Docker setups, and as such only
                # accessible through this environmental variable
                attempt_count = 0
                while attempt_count < num_tries:
                    try:
                        res = self.session.send(req, stream=stream, timeout=timeout)
                        break
                    except Timeout:
                        # Retrying due to a failure of some kind
                        attempt_count += 1
                        # Exponential sleep between calls (up to 10 seconds)
                        time.sleep(min(pow(2, attempt_count), 10))
                if res is None:
                    error_msg = (
                        'API call failed due to unknown reason. '
                        'Please try again at a later point.'
                    )
                    raise Timeout(error_msg)
            else:
                res = self.session.send(req, stream=stream)

            if self.verbose:
                assert res is not None, 'res unexpectedly None'
                LOG.info(f'response: {res.text}')

        # catch any failure
        assert res is not None, 'res unexpectedly None'
        if res.status_code != 200:
            self._handle_fail_res(res, endpoint, path[0])

        if stream:
            return self._process_streaming_call_result(res, endpoint, raw_request_info)
        return self._process_non_streaming_call_result(res, endpoint, raw_request_info)

    @staticmethod
    def _raise_on_status_error(
        response: requests.Response, endpoint: str, raw_request_info: str
    ):
        """Raises exception on HTTP errors similar to
        `response.raise_for_status()`."""
        # catch non-auth failures
        try:
            response.raise_for_status()
        except Exception:
            response_payload = response.json()
            try:
                failure_message = response_payload.get('message', 'Unknown')
                failure_stacktrace = response_payload.get('stacktrace')
                error_msg = (
                    f'API call failed.\n'
                    f'Error message: {failure_message}\n'
                    f'Endpoint: {endpoint}'
                )
                if failure_stacktrace:
                    error_msg += f'\nStacktrace: {failure_stacktrace}'

            except KeyError:
                error_msg = (
                    f'API call to {endpoint} failed.\n'
                    f'Request response: {response.text}'
                )
            LOG.debug(f'{error_msg}\n{raw_request_info}')
            raise RuntimeError(error_msg)

    def _process_non_streaming_call_result(
        self, response: requests.Response, endpoint: str, raw_request_info: str
    ):

        Connection._raise_on_status_error(response, endpoint, raw_request_info)

        # catch non-JSON response (this is rare, the backend should generally
        # return JSON in all cases)
        try:
            response_payload = response.json()
        except json.JSONDecodeError:
            error_msg = (
                f'API call to {endpoint} failed.\n' f'Request response: {response.text}'
            )
            LOG.exception(f'{response.status_code}\nf{error_msg}')

            LOG.debug(f'{error_msg}\n{raw_request_info}')
            raise RuntimeError(error_msg)

        assert response_payload['status'] == self.SUCCESS_STATUS
        result = response_payload.get('result')
        self.last_server_log = response_payload.get('logs')

        # log the API call on success (excerpt response on success)
        response_excerpt = textwrap.indent(
            json.dumps(response_payload, indent=2)[:2048], '  '
        )
        log_msg = (
            f'API call to {endpoint} succeeded.\n'
            f'Request response: {response_excerpt}\n'
            f'{raw_request_info}\n'
        )
        if self.verbose:
            LOG.info(log_msg)
        return result

    @staticmethod
    def _process_streaming_call_result(
        response: requests.Response, endpoint: str, raw_request_info: str
    ):
        """Processes response in jsonlines format. `json_streaming_endpoint`
        returns jsonlines with one json object per line when
        'X-Fiddler-Response-Format' header is set to 'jsonlines'.
        :returns: a generator for results."""

        Connection._raise_on_status_error(response, endpoint, raw_request_info)

        got_eos = False  # got proper end_of_stream.

        if response.headers.get('Content-Type') != 'application/x-ndjson':
            RuntimeError('Streaming response Content-Type is not "x-ndjson"')

        # Read one line at a time. `chunk_size` None ensures that a line
        # is returned as soon as it is read, rather waiting for any minimum
        # size (default is 512 bytes).
        for line in response.iter_lines(chunk_size=None):
            if line:
                try:
                    row_json = json.loads(line)
                    if 'result' in row_json:
                        yield row_json['result']
                    elif row_json.get('status') == Connection.SUCCESS_STATUS:
                        got_eos = True
                        break
                    elif row_json.get('status') == Connection.FAILURE_STATUS:
                        raise Exception(row_json.get('message'))
                except ValueError as valueError:
                    failure_response = {}
                    failure_response['status'] = 400
                    failure_response['message'] = 'Unable to serialize {line}'
                    failure_response['stacktrace'] = valueError
                    Connection._handle_fail_res(failure_response, '', '')

        if not got_eos:
            raise RuntimeError(
                'Truncated response for streaming request. '
                'Failed to receive successful status.'
            )

    def call_executor_service(
        self,
        path: List[str],
        json_payload: Any = None,
        files: Optional[List[Path]] = None,
        is_get_request: bool = False,
        stream: bool = False,
    ):
        no_auth_onebox = False
        try:
            if self.url == 'http://localhost:6100':
                no_auth_onebox = True
                self.url = 'http://localhost:5100'

            return self.call(path, json_payload, files, is_get_request, stream)
        finally:
            if no_auth_onebox:
                self.url = 'http://localhost:6100'



import logging
import os
import shutil
import tempfile
import zipfile
from abc import ABC, abstractmethod
from io import BytesIO
from itertools import tee

from fastavro import block_reader as AvroReader
from werkzeug.datastructures import FileStorage

from fiddler.file_processor.src.constants import (
    SENTINEL_ABSENT,
    SUPPORTABLE_FILE_EXTENSIONS,
    FiddlerPublishSchema,
)

from .extractor import FileExtractor
from .utils import open_file_w_auto_close

# from file_processor_utils import getProgressBar
LOG = logging.getLogger(__name__)

DICT_SEPARATOR = '/'
LIST_INDEXER_OPENER = '['
LIST_INDEXER_CLOSER = ']'


class FileProcessorBase(ABC):
    """
    FileProcessorBase is an abstract class. If we need to support any new file type. We need to override the validate
    and transform functions, which defines how we want to convert a given file into desired output type. For now, we are
    converting all the records into python dictionary and then to CSV.
    """

    @staticmethod
    def get_processor(file_type, schema):
        if file_type == 'csv':
            return CsvFileProcessor()
        elif file_type == 'avro':
            return AvroFileProcessor(schema)
        raise ValueError(
            f'Invalid file type. Supported file types are : {SUPPORTABLE_FILE_EXTENSIONS}'
        )

    def process(self, file_list) -> None:
        # file_list = self.unzip(file_list)
        file_list = self.validate(file_list)
        return self.transform(file_list)

    def unzip(self, file_list):
        unzipped_files_list = []
        old_temp_dir = ''

        # Extract files to new temporary directory.
        new_temp_dir = tempfile.TemporaryDirectory()
        for temp_dir, file_path, file_extension in file_list:
            old_temp_dir = temp_dir

            if zipfile.is_zipfile(file_path):
                zip_ref = zipfile.ZipFile(file_path)
                zip_ref.extractall(new_temp_dir.name)
                file_lists = os.listdir(new_temp_dir.name)

                # Generate list of files with new_temp_dir, file, and extension.
                for file_name in file_lists:
                    file_extension = FileExtractor.get_supportable_file_extension(
                        file_name
                    )
                    file_path = os.path.join(new_temp_dir.name, file_name)
                    if file_extension is not None:
                        with open(file_path, 'rb') as fh:
                            buf = BytesIO(fh.read())
                            file = {file_name: FileStorage(buf, file_name)}
                        unzipped_files_list.append((new_temp_dir, file, file_extension))
                zip_ref.close()

        # Clean old temporary directory
        try:
            shutil.rmtree(old_temp_dir.name)
        except (FileNotFoundError, Exception, shutil.Error):
            LOG.info(f'failed to delete directory : {old_temp_dir}')

        return unzipped_files_list

    def parse_key(self, d, k):
        """
        Will properly parse and handle nested values indicated as DICT_SEPARATOR

        # value/nested/final
        # list[-1]
        # list[-1][-1]
        # value/nested/list[-1]/final
        """
        DEFAULT_VALUE = SENTINEL_ABSENT
        ABSENT_IND_VALUE = len(k) + 1

        list_ind_opener = (
            k.find(LIST_INDEXER_OPENER)
            if k.find(LIST_INDEXER_OPENER) > -1
            else ABSENT_IND_VALUE
        )
        dict_sep_ind = (
            k.find(DICT_SEPARATOR) if k.find(DICT_SEPARATOR) > -1 else ABSENT_IND_VALUE
        )

        if k == '':
            # List ends value
            return d

        if dict_sep_ind == list_ind_opener == ABSENT_IND_VALUE:
            # No nested value, no lists. Easy
            return d.get(k, DEFAULT_VALUE)

        if dict_sep_ind < list_ind_opener:
            # Dict separator comes first (or list separator is absent)
            dict_sep_ind = k.find(DICT_SEPARATOR)
            sub_key = k[:dict_sep_ind]

            nested_d = d.get(sub_key, DEFAULT_VALUE)
            if nested_d == DEFAULT_VALUE or nested_d is None:
                return DEFAULT_VALUE

            return self.parse_key(nested_d, k[dict_sep_ind + 1 :])

        elif dict_sep_ind > list_ind_opener:
            # List separator comes first (or dict separator is absent)
            while True:
                opener = k.find(LIST_INDEXER_OPENER)
                closer = k.find(LIST_INDEXER_CLOSER)

                if opener != 0:
                    ind = int(k[opener + 1 : closer])
                    sub_key = k[:opener]
                    nested_d = d.get(sub_key, DEFAULT_VALUE)
                    if nested_d == DEFAULT_VALUE or nested_d is None:
                        return DEFAULT_VALUE

                    d = d[sub_key][ind]
                    k = k[closer + 1 :]
                else:
                    ind = int(k[opener + 1 : closer])
                    d = d[ind]
                    k = k[closer + 1 :]

                if len(k) > 0 and k[0] == '/':
                    # Edge case, next key is a sub
                    k = k[1:]

                if len(k) > 0 and k[0] != LIST_INDEXER_OPENER or len(k) == 0:
                    break

            return self.parse_key(d, k)

        else:
            return DEFAULT_VALUE

    def generate_events(
        self, publish_schema, parent_event_template, rows_iterator, results=[]
    ):
        """
        Transforms the passed event by fixing field names to be DB friendly,
        and then removing fields that are not part of `supported_fields` or a timestamp_column
        """
        event_template = parent_event_template.copy()

        # Get static info
        if FiddlerPublishSchema.STATIC in publish_schema:
            for f_mapping, static_value in publish_schema[
                FiddlerPublishSchema.STATIC
            ].items():
                if static_value is SENTINEL_ABSENT:
                    continue

        for row in rows_iterator:
            row_template = event_template.copy()

            # Get dynamic info
            if FiddlerPublishSchema.DYNAMIC in publish_schema:
                for f_mapping, d_mapping in publish_schema[
                    FiddlerPublishSchema.DYNAMIC
                ].items():
                    if d_mapping is SENTINEL_ABSENT:
                        continue

                    cleaned_val = self.parse_key(row, d_mapping)
                    if cleaned_val is SENTINEL_ABSENT:
                        continue

                    row_template[f_mapping] = cleaned_val

            # Check for iterator recursion point in AVRO files
            if FiddlerPublishSchema.ITERATOR in publish_schema:
                if isinstance(publish_schema[FiddlerPublishSchema.ITERATOR], list):
                    recursed_successfully = False

                    for iter_schema in publish_schema[FiddlerPublishSchema.ITERATOR]:
                        curr_iterator_key = iter_schema[
                            FiddlerPublishSchema.ITERATOR_KEY
                        ]
                        row_iterator = self.parse_key(row, curr_iterator_key)
                        if row_iterator is SENTINEL_ABSENT:
                            # Default behavior if no row iterator present for that key: Pass
                            #     This is done to prevent multiple publishing of exact same base
                            continue

                        recursed_successfully = True
                        self.generate_events(
                            iter_schema, row_template, row_iterator, results
                        )

                    if not recursed_successfully:
                        # Default behavior if no row iterator present for ANY key: publish at current level
                        # log(f'Iterators {iter_keys} specified, but no iterators found in row. Default publishing event at current level.')
                        # publish(event_template)
                        return

                elif isinstance(publish_schema[FiddlerPublishSchema.ITERATOR], dict):
                    iter_schema = publish_schema[FiddlerPublishSchema.ITERATOR]
                    curr_iterator_key = iter_schema[FiddlerPublishSchema.ITERATOR_KEY]
                    row_iterator = self.parse_key(row, curr_iterator_key)
                    if row_iterator is SENTINEL_ABSENT:
                        # Default behavior if no row iterator present for that key: publish at current level
                        # publish(event_template)
                        return
                    else:
                        # print(f' row iterator before result :{row_iterator}')
                        self.generate_events(
                            iter_schema, row_template, row_iterator, results
                        )
                else:
                    # Iterator present in config, but not in row. Default behavior: Publish as is.
                    # publish(event_template)
                    results.append(row_template)
            else:
                # No iterator, publish as is
                # publish(event_template)
                results.append(row_template)
        return results

    @abstractmethod
    def validate(self, file_list):
        pass

    @abstractmethod
    def transform(self, file_list):
        pass


class CsvFileProcessor(FileProcessorBase):
    def validate(self, file_list):
        pass

    def transform(self, file_list):
        pass


class AvroFileProcessor(FileProcessorBase):
    def __init__(self, schema):
        if schema is None:
            raise ValueError('Schema is required for AVRO files')
        self.schema = schema

    def validate(self, file_list):
        LOG.info(f'publish_schema :{self.schema}')
        publish_schema = self.schema
        invalid_keys = self.validate_schema(publish_schema, set())
        if len(invalid_keys) > 0:
            raise ValueError(
                f'Invalid Schema. Schema contains invalid keys : {invalid_keys}.\
                            Valid keys are : {FiddlerPublishSchema.VALID_SCHEMA_KEYWORDS}'
            )
        return file_list

    def validate_schema(self, schema_json, invalid_keys):
        for key in schema_json:
            if key in FiddlerPublishSchema.VALID_SCHEMA_KEYWORDS:
                if key is FiddlerPublishSchema.ITERATOR:
                    return self.validate_schema(schema_json[key], invalid_keys)
            else:
                invalid_keys.add(key)
        return invalid_keys

    def transform(self, file_list):
        results = []
        for temp_path, file_path, file_extension in file_list:
            file_path_str = file_path if isinstance(file_path, str) else file_path.name
            LOG.info(f'avro file path {file_path_str}')
            events_collection = AvroReader(open(file_path_str, 'rb'))
            results.extend(self.process_avro(events_collection))
        return results

    def process_avro(self, avro_reader):
        """
        Breaks the inputstream into a series of rows.

        """
        avro_reader, block_iter = tee(avro_reader)
        num_rows = 0
        for block in block_iter:
            num_rows += block.num_records

        results = []
        for batch, block in enumerate(avro_reader):
            results = self.generate_events(
                publish_schema=self.schema,
                parent_event_template={},
                rows_iterator=block,
                results=results,
            )
            results = [i for i in results if i]
        return results



import enum
from typing import Any, List, Optional

from pydantic import Field

from fiddler.v2.schema.base import BaseDataSchema


@enum.unique
class DataType(str, enum.Enum):
    """Supported datatypes for the Fiddler engine."""

    FLOAT = 'float'
    INTEGER = 'int'
    BOOLEAN = 'bool'
    STRING = 'str'
    CATEGORY = 'category'

    def is_numeric(self):
        return self.value in (DataType.INTEGER.value, DataType.FLOAT.value)

    def is_bool_or_cat(self):
        return self.value in (DataType.BOOLEAN.value, DataType.CATEGORY.value)

    def is_valid_target(self):
        return self.value != DataType.STRING.value


class Column(BaseDataSchema):
    """Represents a single column of a dataset or model input/output.

    :param name: The name of the column (corresponds to the header row of a
        CSV file)
    :param data_type: The best encoding type for this column's data.
    :param possible_values: If data_type is CATEGORY, then an exhaustive list
        of possible values for this category must be provided. Otherwise
        this field has no effect and is optional.
    :param is_nullable: Optional metadata. Tracks whether or not this column is
        expected to contain some null values.
    :param value_range_x: Optional metadata. If data_type is FLOAT or INTEGER,
        then these values specify a range this column's values are expected to
        stay within. Has no effect for non-numerical data_types.
    """

    name: str = Field(alias='column-name')
    data_type: DataType = Field(alias='data-type')
    possible_values: Optional[List[Any]] = Field(None, alias='possible-values')
    is_nullable: Optional[bool] = Field(None, alias='is-nullable')
    value_range_min: Optional[float] = Field(None, alias='value-range-min')
    value_range_max: Optional[float] = Field(None, alias='value-range-max')

    class Config:
        use_enum_values = True
        allow_population_by_field_name = True

    # # @TODO: Use pydantic validator
    # inappropriate_value_range = not self.data_type.is_numeric() and not (
    #     self.value_range_min is None and self.value_range_max is None
    # )
    # if inappropriate_value_range:
    #     raise ValueError(
    #         f'Do not pass `value_range` for '
    #         f'non-numerical {self.data_type} data type.'
    #     )

    @classmethod
    def from_dict(cls, desrialized_json: dict):
        return cls.parse_obj(desrialized_json)


class DatasetInfo(BaseDataSchema):
    columns: List[Column]



import json
import tempfile
from http import HTTPStatus
from pathlib import Path
from typing import Any, Dict, Optional, Union
from fiddler.core_objects import BatchPublishType

import pandas as pd
from requests_toolbelt.multipart.encoder import MultipartEncoder

from fiddler.libs.http_client import APP_JSON_CONTENT_TYPE, RequestClient
from fiddler.utils import logging
from fiddler.v2.api.helpers import multipart_upload
from fiddler.v2.constants import FiddlerTimestamp, FileType, UploadType
from fiddler.v2.schema.events import EventIngest, EventsIngest
from fiddler.v2.utils.exceptions import FiddlerAPIException, handle_api_error_response
from fiddler.v2.utils.response_handler import APIResponseHandler

logger = logging.getLogger(__name__)


class EventsMixin:
    client: RequestClient
    organization_name: str

    @handle_api_error_response
    def publish_events_batch(
        self,
        project_name: str,
        model_name: str,
        events_path: Path,
        batch_id: Optional[str] = None,
        events_schema: Optional[str] = None,
        id_field: Optional[str] = None,
        is_update: Optional[bool] = None,
        timestamp_field: Optional[str] = None,
        timestamp_format: Optional[FiddlerTimestamp] = None,
        group_by: Optional[str] = None,
        file_type: Optional[FileType] = None,
        is_sync: Optional[bool] = True,
    ) -> Dict[str, str]:
        """
        Publishes a batch events object to Fiddler Service.

        :param project_name: The project to which the model whose events are being published belongs.
        :param model_name: The model whose events are being published.
        :param events_path: pathlib.Path pointing to the events file to be uploaded
        :param batch_id: <TBD>
        :param events_schema: <TBD>
        :param id_field: Column to extract id value from.
        :param is_update: Bool indicating if the events are updates to previously published rows
        :param timestamp_field: Column to extract timestamp value from.
                                Timestamp must match the specified format in `timestamp_format`.
        :param timestamp_format:Format of timestamp within batch object. Can be one of:
                                - FiddlerTimestamp.INFER
                                - FiddlerTimestamp.EPOCH_MILLISECONDS
                                - FiddlerTimestamp.EPOCH_SECONDS
                                - FiddlerTimestamp.ISO_8601
        :param group_by: Column to group events together for Model Performance metrics. For example,
                        in ranking models that column should be query_id or session_id, used to
                        compute NDCG and MAP. Be aware that the batch_source file/dataset provided should have
                        events belonging to the SAME query_id/session_id TOGETHER and cannot be mixed
                        in the file. For example, having a file with rows belonging to query_id 31,31,31,2,2,31,31,31
                        would not work. Please sort the file by group_by group first to have rows with
                        the following order: query_id 31,31,31,31,31,31,2,2.
        :param file_type: FileType which specifices the filetype csv etc.
        :param is_sync: A boolean value which determines if the upload method works in synchronous mode or async mode
        :returns: Dictionary containing details of the job used to publish events incase of 202 response from the server.
        """
        file_name = events_path.name
        response = multipart_upload(
            client=self.client,
            organization_name=self.organization_name,
            project_name=project_name,
            identifier=model_name,
            upload_type=UploadType.EVENT.value,
            file_path=str(events_path),
            file_name=file_name,
        )
        file_name = response.get('file_name')
        request_body = EventsIngest(
            batch_id=batch_id,
            events_schema=events_schema,
            id_field=id_field,
            is_update=is_update,
            timestamp_field=timestamp_field,
            timestamp_format=timestamp_format,
            group_by=group_by,
            file_type=file_type,
            file_name=[file_name],
        ).dict()
        response = self.client.post(
            url=f'events/{self.organization_name}:{project_name}:{model_name}/ingest',
            data=request_body,
        )
        # @TODO: Handle invalid file path exception
        if response.status_code == HTTPStatus.ACCEPTED:
            resp = APIResponseHandler(response).get_data()
            if is_sync:
                job_id = resp['job_uuid']
                job = self.poll_job(job_id).get_data()
                job.pop('extras', None)
                return job
            else:
                return resp
        else:
            # raising a generic FiddlerAPIException
            logger.error('Failed to publish events')
            raise FiddlerAPIException(
                response.status_code,
                error_code=response.status_code,
                message=response.content,
                errors=[],
            )

    @handle_api_error_response
    def publish_event(
        self,
        project_name: str,
        model_name: str,
        event: dict,
        event_id: Optional[str] = None,
        id_field: Optional[str] = None,
        is_update: Optional[bool] = None,
        event_timestamp: Optional[str] = None,
        timestamp_format: Optional[str] = None,
    ) -> Optional[str]:
        """
        Publishes an event to Fiddler Service.

        :param project_name: The project to which the model whose events are being published belongs
        :param model_name: The model whose events are being published
        :param dict event: Dictionary of event details, such as features and predictions.
        :param event_id: Unique str event id for the event
        :param event_timestamp: The UTC timestamp of the event in epoch milliseconds (e.g. 1609462800000)
        :param timestamp_format: Format of timestamp within batch object. Can be one of:
                                - FiddlerTimestamp.INFER
                                - FiddlerTimestamp.EPOCH_MILLISECONDS
                                - FiddlerTimestamp.EPOCH_SECONDS
                                - FiddlerTimestamp.ISO_8601
        :returns: Unique event id incase of successful submitted request.
        """
        request_body = EventIngest(
            event=event,
            event_id=event_id,
            id_field=id_field,
            is_update=is_update,
            event_timestamp=event_timestamp,
            timestamp_format=timestamp_format,
        ).dict()
        response = self.client.post(
            url=f'events/{self.organization_name}:{project_name}:{model_name}/ingest/event',
            data=request_body,
        )
        if response.status_code == HTTPStatus.ACCEPTED:
            response_dict = APIResponseHandler(response).get_data()
            logger.info(response_dict.get('message'))
            return response_dict.get('__fiddler_id')
        else:
            # raising a generic FiddlerAPIException
            logger.error('Failed to publish events')
            raise FiddlerAPIException(
                response.status_code,
                error_code=response.status_code,
                message=response.content,
                errors=[],
            )

    @handle_api_error_response
    def publish_events_batch_dataframe(
        self,
        project_name: str,
        model_name: str,
        events_df: pd.DataFrame,
        batch_id: Optional[str] = None,
        id_field: Optional[str] = None,
        is_update: Optional[bool] = None,
        timestamp_field: Optional[str] = None,
        timestamp_format: Optional[FiddlerTimestamp] = None,
        group_by: Optional[str] = None,
        is_sync: Optional[bool] = True,
    ) -> Dict[str, str]:
        """
        Publishes a batch events object to Fiddler Service.

        :param project_name: The project to which the model whose events are being published belongs.
        :param model_name: The model whose events are being published.
        :param events_df: pd.DataFrame object having the events
        :param batch_id: <TBD>
        :param id_field: Column to extract id value from.
        :param is_update: Bool indicating if the events are updates to previously published rows
        :param timestamp_field: Column to extract timestamp value from.
                                Timestamp must match the specified format in `timestamp_format`.
        :param timestamp_format: Format of timestamp within batch object. Can be one of:
                                - FiddlerTimestamp.INFER
                                - FiddlerTimestamp.EPOCH_MILLISECONDS
                                - FiddlerTimestamp.EPOCH_SECONDS
                                - FiddlerTimestamp.ISO_8601
        :param group_by: Column to group events together for Model Performance metrics. For example,
                        in ranking models that column should be query_id or session_id, used to
                        compute NDCG and MAP. Be aware that the batch_source file/dataset provided should have
                        events belonging to the SAME query_id/session_id TOGETHER and cannot be mixed
                        in the file. For example, having a file with rows belonging to query_id 31,31,31,2,2,31,31,31
                        would not work. Please sort the file by group_by group first to have rows with
                        the following order: query_id 31,31,31,31,31,31,2,2.
        :param is_sync: A boolean value which determines if the upload method works in synchronous mode or async mode
        :returns: Dictionary containing details of the job used to publish events incase of 202 response from the server.
        """
        file_type = FileType.CSV
        with tempfile.NamedTemporaryFile(suffix=file_type) as temp:
            events_df.to_csv(temp, index=False)
            events_path = Path(temp.name)
            return self.publish_events_batch(
                project_name=project_name,
                model_name=model_name,
                events_path=events_path,
                batch_id=batch_id,
                id_field=id_field,
                is_update=is_update,
                timestamp_field=timestamp_field,
                timestamp_format=timestamp_format,
                group_by=group_by,
                is_sync=is_sync,
            )

    @handle_api_error_response
    def publish_events_batch_schema(  # noqa
        self,
        events_path: Union[Path, str],
        publish_schema: Dict[str, Any],
        data_source: Optional[BatchPublishType] = None,
        is_sync: Optional[bool] = True,
    ) -> Dict[str, str]:
        """
        Publishes a batch events object to Fiddler Service.
        :param events_path: pathlib.Path pointing to the events file to be uploaded
        :param publish_schema: Dict object specifying layout of data.
        :param is_sync: A boolean value which determines if the upload method works in synchronous mode or async mode
        :returns: Dictionary containing details of the job used to publish events incase of 202 response from the server.
        """
        request_body_json = {}
        request_body_json['schema'] = publish_schema
        if data_source == BatchPublishType.AWS_S3 and isinstance(events_path, str):
            request_body = {
                's3_url': events_path,
                'schema': json.dumps(publish_schema)
            }
            response = self.client.post(
                url=f'events/{self.organization_name}/ingest/schema',
                headers={'Content-Type': 'application/json'},
                data=request_body,
            )
        else:
            file_name = events_path.name
            files = {}
            files[file_name] = events_path
            request_body = {
                file_path.name: (
                    file_path.name,
                    open(file_path, 'rb'),
                )
                for _, file_path in files.items()
            }
            # https://stackoverflow.com/a/19105672/13201804
            request_body.update(
                {'schema': (None, json.dumps(publish_schema), 'application/json')}
            )
            m = MultipartEncoder(fields=request_body)
            content_type_header, request_body  = m.content_type, m
            # content_type_header, request_body  = event_ingest.multipart_form_request()
            response = self.client.post(
                url=f'events/{self.organization_name}/ingest/schema',
                headers={'Content-Type': content_type_header},
                data=request_body,
            )
        # @TODO: Handle invalid file path exception
        if response.status_code == HTTPStatus.ACCEPTED:
            resp = APIResponseHandler(response).get_data()
            if is_sync:
                job_id = resp['job_uuid']
                job = self.poll_job(job_id).get_data()
                job.pop('extras', None)
                return job
            else:
                return resp

        else:
            # raising a generic FiddlerAPIException
            logger.error('Failed to upload events schema.')
            raise FiddlerAPIException(
                response.status_code,
                error_code=response.status_code,
                message=response.content,
                errors=[],
            )



"""
Fiddler Client Module
=====================

A Python client for Fiddler service.

TODO: Add Licence.
"""
import configparser
import functools
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from fiddler.api.monitoring import Monitoring
from fiddler.model import Model

from . import utils
from ._version import __version__
from .client import Fiddler, PredictionEventBundle
from .core_objects import (
    BatchPublishType,
    Column,
    CustomFeature,
    DatasetInfo,
    DataType,
    DeploymentOptions,
    DeploymentType,
    ExplanationMethod,
    FiddlerPublishSchema,
    FiddlerTimestamp,
    MLFlowParams,
    ModelDeploymentParams,
    ModelInfo,
    ModelInputType,
    ModelTask,
    WeightingParams,
)
from .fiddler_api import FiddlerApi as FiddlerApiV1
from .file_processor.src.constants import (
    CSV_EXTENSION,
    PARQUET_COMPRESSION,
    PARQUET_ENGINE,
    PARQUET_EXTENSION,
    PARQUET_ROW_GROUP_SIZE,
)
from .packtools import gem
from .utils import ColorLogger
from .v1_v2_compat import V1V2Compat
from .v2.api.api import Client as FiddlerApiV2
from .v2.constants import ServerDeploymentMode
from .v2.schema.job import JobStatus
from .validator import PackageValidator, ValidationChainSettings, ValidationModule

logger = utils.logging.getLogger(__name__)

VERSIONS = [1, 2]


class FiddlerApi:

    """Client of all connections to the Fiddler API.
    :param url:         The base URL of the API to connect to. Usually either
        https://<yourorg>.fiddler.ai (cloud) or http://localhost:4100 (onebox)
    :param org_id:      The name of your organization in the Fiddler platform
    :param auth_token:  Token used to authenticate. Your token can be
        created, found, and changed at <FIDDLER URL>/settings/credentials.
    :param proxies:     Optionally, a dict of proxy URLs. e.g.,
                    proxies = {'http' : 'http://proxy.example.com:1234',
                               'https': 'https://proxy.example.com:5678'}
    :param verbose:     if True, api calls will be logged verbosely,
                    *warning: all information required for debugging will be
                    logged including the auth_token.
    :param timeout:     How long to wait for the server to respond before giving up
    :param version:     Version of the client you want to instantiate. Options [1,2]
    """

    def __new__(
        cls,
        url: Optional[str] = None,
        org_id: Optional[str] = None,
        auth_token: Optional[str] = None,
        proxies: Optional[dict] = None,
        verbose: Optional[bool] = False,
        timeout: int = 1200,  # sec
        version: int = 1,
    ):
        url, org_id, auth_token = cls._get_connection_parameters(
            cls, url, org_id, auth_token, version
        )

        # Validation of version, org_id is handled by FiddlerApiV1.
        client_v1 = FiddlerApiV1(
            url=url,
            org_id=org_id,
            auth_token=auth_token,
            proxies=proxies,
            verbose=verbose,
            timeout=timeout,
        )
        # @todo: Handle proxies in v2
        client_v2 = FiddlerApiV2(
            url=url, organization_name=org_id, auth_token=auth_token, timeout=timeout
        )
        supported_features = cls._get_supported_features(cls, client_v1, org_id)
        server_deployment_mode = cls._get_server_deployment_mode(
            cls, supported_features
        )

        logger.info(f'Version deployed on the server side {server_deployment_mode}')

        compat_client = V1V2Compat(client_v2)

        if version == 2:
            logger.info('Default client is of client version 2')

            obj = lambda: None  # instantiate an empty object # noqa

            obj.list_datasets = compat_client.get_datasets
            obj.get_dataset_info = compat_client.get_dataset_info
            obj.get_dataset = compat_client.get_dataset_artifact
            obj.delete_dataset = compat_client.delete_dataset

            obj.list_projects = compat_client.get_projects
            obj.create_project = compat_client.add_project
            obj.delete_project = compat_client.delete_project

            obj.upload_dataset = (
                compat_client.upload_dataset_dataframe
            )  # this is uploading dataframe in v1 and csv in v2
            obj.upload_dataset_from_file = (
                compat_client.upload_dataset
            )  # currently only supports csv
            obj.upload_dataset_from_dir = compat_client.upload_dataset_from_dir
            # obj.process_csv = client_v1.process_csv
            # obj.process_avro = client_v1.process_avro

            obj.publish_event = compat_client.publish_event
            obj.publish_events_batch = compat_client.publish_events_batch
            # obj.publish_events_log = client_v1.publish_events_log #deprecated so not implemented
            obj.publish_events_batch_schema = compat_client.publish_events_batch_schema
            # obj.publish_parquet_s3 = client_v1.publish_parquet_s3
            obj.generate_sample_events = client_v1.generate_sample_events

            obj.upload_model_package = functools.partial(
                functools.partial(v1_upload_model_package, v1_client=client_v1),
                v2_client=client_v2,
            )

            obj.trigger_pre_computation = client_v1.trigger_pre_computation

            obj.register_model = functools.partial(
                functools.partial(v1_register_model, v1_client=client_v1),
                v2_client=client_v2,
            )
            obj.update_model = client_v1.update_model
            obj.add_model = functools.partial(
                functools.partial(add_model, v1_client=client_v1),
                v2_client=client_v2,
            )

            obj._trigger_model_predictions = client_v1._trigger_model_predictions

            # explictly binding non-conflicting v1 methods to the v2 obj
            # conflicting methods will use v2 method by default for this condition
            # obj.project = client_v1.project

            obj.list_models = client_v1.list_models
            obj.get_model_info = client_v1.get_model_info
            obj.delete_model = client_v1.delete_model

            obj.get_slice = client_v1.get_slice

            obj.run_model = client_v1.run_model
            obj.run_explanation = client_v1.run_explanation
            obj.run_feature_importance = client_v1.run_feature_importance
            obj.run_fairness = client_v1.run_fairness
            obj.get_mutual_information = client_v1.get_mutual_information

            # The below methods are not used so not mapping them.
            # obj.share_project = client_v1.share_project
            # obj.unshare_project = client_v1.unshare_project
            # obj.list_org_roles = client_v1.list_org_roles
            # obj.list_project_roles = client_v1.list_project_roles
            # obj.list_teams = client_v1.list_teams

            # @todo make this available under client.v2
            # obj.initialize_monitoring = functools.partial(
            #     obj._initialize_monitoring, v1_connection=client_v1.connection
            # )

            # obj.add_monitoring_config = client_v1.add_monitoring_config
            # obj.list_segments = client_v1.list_segments
            # obj.upload_segment = client_v1.upload_segment
            # obj.activate_segment = client_v1.activate_segment
            # obj.deactivate_segment = client_v1.deactivate_segment
            # obj.delete_segment = client_v1.delete_segment
            # obj.get_segment_info = client_v1.get_segment_info
            # also retaining the sub-module approach
        else:
            logger.info('Default client is of client version 1')
            obj = client_v1
            if ServerDeploymentMode.F2 == server_deployment_mode:
                obj.publish_event = compat_client.publish_event
                obj.publish_events_batch = compat_client.publish_events_batch
                # obj.publish_events_batch_schema

        obj.v1 = client_v1
        obj.v2 = client_v2
        return obj

    def _get_connection_parameters(
        self, url: str, org_id: str, auth_token: str, version: int
    ) -> Tuple:
        if Path('fiddler.ini').is_file():
            config = configparser.ConfigParser()
            config.read('fiddler.ini')
            info = config['FIDDLER']
            if not url:
                url = info.get('url', None)
            if not org_id:
                org_id = info.get('org_id', None)
            if not auth_token:
                auth_token = info.get('auth_token', None)

        if not url:
            raise ValueError('Could not find url. Please enter a valid url')
        if not org_id:
            raise ValueError('Could not find org_id. Please enter a valid org_id')
        if not auth_token:
            raise ValueError(
                'Could not find auth_token. Please enter a valid auth_token'
            )
        if version not in VERSIONS:
            raise ValueError(
                f'version={version} not supported. Please enter a valid version. Supported versions={VERSIONS}'
            )

        return url, org_id, auth_token

    def _get_supported_features(self, client_v1: FiddlerApiV1, org_id: str) -> Dict:
        supported_features: Dict = None

        try:
            path: List['str'] = ['get_supported_features', org_id]
            supported_features = client_v1.connection.call(path, is_get_request=True)
        except Exception as e:
            logger.warning('Fetching supported features failed with error ' + e.msg)

        return supported_features

    def _get_server_deployment_mode(
        self, supported_features: Optional[Dict] = None
    ) -> ServerDeploymentMode:
        ret_val: ServerDeploymentMode = ServerDeploymentMode.F1

        if supported_features:
            ret_val = (
                ServerDeploymentMode.F2
                if supported_features.get('enable_fiddler_v2', False)
                else ServerDeploymentMode.F1
            )

        return ret_val


def v1_upload_model_package(
    artifact_path: Path,
    project_id: str,
    model_id: str,
    deployment_type: Optional[
        str
    ] = DeploymentType.PREDICTOR,  # model deployment type. One of {'predictor', 'executor'}
    image_uri: Optional[str] = None,  # image to be used for newly uploaded model
    namespace: Optional[str] = None,  # kubernetes namespace
    port: Optional[int] = 5100,  # port on which model is served
    replicas: Optional[int] = 1,  # number of replicas
    cpus: Optional[float] = 0.25,  # number of CPU cores
    memory: Optional[str] = '128m',  # amount of memory required.
    gpus: Optional[int] = 0,  # number of GPU cores
    await_deployment: Optional[bool] = True,  # wait for deployment
    is_sync=True,
    v2_client: FiddlerApiV2 = None,
    v1_client: FiddlerApiV1 = None,
):
    v1_client.upload_model_package(
        artifact_path,
        project_id,
        model_id,
        deployment_type,
        image_uri,
        namespace,
        port,
        replicas,
        cpus,
        memory,
        gpus,
        await_deployment,
    )

    call_init_monitoring(v1_client, v2_client, project_id, model_id, is_sync)


def v1_register_model(
    project_id: str,
    model_id: str,
    dataset_id: str,
    model_info: ModelInfo,
    deployment: Optional[DeploymentOptions] = None,
    cache_global_impact_importance: bool = True,
    cache_global_pdps: bool = False,
    cache_dataset: bool = True,
    is_sync=True,
    v2_client: FiddlerApiV2 = None,
    v1_client: FiddlerApiV1 = None,
):
    v1_client.register_model(
        project_id,
        model_id,
        dataset_id,
        model_info,
        deployment,
        cache_global_impact_importance,
        cache_global_pdps,
        cache_dataset,
    )
    call_init_monitoring(v1_client, v2_client, project_id, model_id, is_sync)


def add_model(
    project_name: str,
    model_name: str,
    model_info: ModelInfo = None,
    model_type: Optional[str] = None,
    framework: Optional[str] = None,
    requirements: Optional[str] = None,
    is_sync: Optional[bool] = True,
    v2_client: FiddlerApiV2 = None,
    v1_client: FiddlerApiV1 = None,
) -> Model:
    """
    Function to add a model to fiddler for monitoring

    :param project_name: project name where the model will be added
    :type project_name: string
    :param model_name: name of the model
    :type model_name: string
    :param info: model related information passed as dictionary from user
    :type info: dictionary
    :param model_type: the type of model e.g. REGRESSION, BINARY CLASSIFICATION
    :type model_type: string
    :return: Model object which contains the model details
    """
    if model_info is None:
        raise ValueError('ModelInfo cannot be None')
    if len(model_info.datasets) != 1:
        raise ValueError('Specify a single dataset for the model')

    dataset_name = model_info.datasets[0]

    dataset = v2_client.get_dataset(
        project_name=project_name, dataset_name=dataset_name
    )
    outputs = model_info.get_output_names()
    dataset_cols = set([col.name for col in dataset.info.columns])

    if not all(elem in dataset_cols for elem in outputs):
        raise ValueError(f'Dataset {dataset_name} does not have output columns')

    info = model_info.to_dict()
    model = v2_client.add_model(
        project_name, model_name, info, model_type, framework, requirements
    )
    call_init_monitoring(v1_client, v2_client, project_name, model_name, is_sync)
    return model


def call_init_monitoring(
    v1_client: FiddlerApiV1,
    v2_client: FiddlerApiV2,
    project_id: str,
    model_id: str,
    is_sync: bool,
):

    logger.info('Start initialize monitoring')
    print('Start initialize monitoring')
    resp = Monitoring(v1_client.connection, project_id, model_id).initialize_monitoring(
        False, False, 'fiddler2'
    )

    if resp and is_sync:
        job_resp_final = v2_client.poll_job(resp)
        if job_resp_final.status == JobStatus.FAILURE:
            err_msg = 'Init monitoring failed with error message'
            logger.error(err_msg)
            raise Exception(f'{err_msg} - {job_resp_final.error_message}')
        elif job_resp_final.status == JobStatus.SUCCESS:
            results = FiddlerApiV2.get_task_results(job_resp_final)
            result_msg = '\n'.join(results)
            success_msg = f'Init monitoring succeeded:{result_msg}'
            logger.info(success_msg)
            print(success_msg)
            return True

        logger.info('Finished initialize monitoring')
        print('Finished initialize monitoring')

    else:
        return resp


__all__ = [
    '__version__',
    'BatchPublishType',
    'Column',
    'CustomFeature',
    'ColorLogger',
    'DatasetInfo',
    'DataType',
    'Fiddler',
    'FiddlerApi',
    'FiddlerTimestamp',
    'FiddlerPublishSchema',
    'gem',
    'MLFlowParams',
    'ModelDeploymentParams',
    'ModelInfo',
    'ModelInputType',
    'ModelTask',
    'ExplanationMethod',
    'PredictionEventBundle',
    'PackageValidator',
    'ValidationChainSettings',
    'ValidationModule',
    'utils',
    # Exposing constants
    'CSV_EXTENSION',
    'PARQUET_EXTENSION',
    'PARQUET_ROW_GROUP_SIZE',
    'PARQUET_ENGINE',
    'PARQUET_COMPRESSION',
]



import json
import os
import unittest

from fiddler.file_processor.src.constants import ExtractorFileType
from fiddler.file_processor.src.processor import AvroFileProcessor

package_dir = os.path.dirname(os.path.abspath(__file__))


class AvroFileProcessorTest(unittest.TestCase):
    def test_validation_with_invalid_schema_should_fail(self):
        # GIVEN
        schema_file_path = os.path.join(
            package_dir, 'resources/avro/simple_avro_invalid_mapping_schema.json'
        )
        with open(schema_file_path) as fs:
            file_schema_str = json.load(fs)
        avro_file_path = os.path.join(package_dir, 'resources/avro/simple_avro.avro')
        files_list = [(package_dir, avro_file_path, ExtractorFileType.AVRO)]
        avro_file_processor = AvroFileProcessor(file_schema_str)

        # WHEN
        self.assertRaises(ValueError, avro_file_processor.process, files_list)

    def test_validation_with_invalid_nested_schema_should_fail(self):
        # GIVEN
        schema_file_path = os.path.join(
            package_dir, 'resources/avro/nested_avro_invalid_mapping_schema.json'
        )
        with open(schema_file_path) as fs:
            file_schema_str = json.load(fs)
        avro_file_path = os.path.join(package_dir, 'resources/avro/simple_avro.avro')
        files_list = [(package_dir, avro_file_path, ExtractorFileType.AVRO)]
        avro_file_processor = AvroFileProcessor(file_schema_str)

        # WHEN
        self.assertRaises(ValueError, avro_file_processor.process, files_list)

    def test_simple_avro_file_should_succeed(self):
        # GIVEN
        schema_file_path = os.path.join(
            package_dir, 'resources/avro/simple_avro_mapping_schema.json'
        )
        with open(schema_file_path) as fs:
            file_schema_str = json.load(fs)
        avro_file_path = os.path.join(package_dir, 'resources/avro/simple_avro.avro')
        files_list = [(package_dir, avro_file_path, ExtractorFileType.AVRO)]
        avro_file_processor = AvroFileProcessor(file_schema_str)

        # WHEN
        result = avro_file_processor.process(files_list)

        # THEN
        expected = [
            {u'station': u'011990-99999', u'temp': 0, u'time': 1433269388},
            {u'station': u'011990-99999', u'temp': 22, u'time': 1433270389},
            {u'station': u'011990-99999', u'temp': -11, u'time': 1433273379},
            {u'station': u'012650-99999', u'temp': 111, u'time': 1433275478},
        ]
        assert len(result) == 4
        assert result == expected

    def test_nested_avro_file_should_succeed(self):
        """
                AVRO RECORD:
                records = [
                {u'name': u'test-account-1', u'transaction':['first_transaction','second_transaction']},
            {u'name': u'test-account-2', u'transaction':['second_first_transaction','second_second_transaction']}
            # {u'name': u'test-account-3'}
        ]
        """
        # GIVEN
        schema_file_path = os.path.join(
            package_dir, 'resources/avro/nested_avro_schema_mapping.json'
        )
        with open(schema_file_path) as fs:
            file_schema_str = json.load(fs)
        avro_file_path = os.path.join(package_dir, 'resources/avro/nested_avro.avro')
        files_list = [(package_dir, avro_file_path, ExtractorFileType.AVRO)]
        avro_file_processor = AvroFileProcessor(file_schema_str)

        # WHEN
        result = avro_file_processor.process(files_list)

        # THEN
        expected = [
            {u'name': u'test-account-1', u'transaction': 'first_transaction'},
            {u'name': u'test-account-1', u'transaction': 'second_transaction'},
            {u'name': u'test-account-2', u'transaction': 'second_first_transaction'},
            {u'name': u'test-account-2', u'transaction': 'second_second_transaction'},
        ]
        assert len(result) == 4
        assert result == expected

    # Commenting as it does not work. I will fix as a part of next revision. Publishing to get early comments on the PR.
    # def test_multi_level_nested_avro_file_should_succeed(self):
    #     # GIVEN
    #     schema_file_path = os.path.join(
    #         package_dir, 'resources/avro/multi_level_nested_avro_schema_mapping.json')
    #     avro_file_path = os.path.join(package_dir, 'resources/avro/multi_level_nested_avro.avro')
    #     files_list = [(package_dir, avro_file_path, ExtractorFileType.AVRO)]
    #     avro_file_processor = AvroFileProcessor(schema_file_path)

    #     # WHEN
    #     result = avro_file_processor.process(files_list)

    #     # THEN
    #     expected = [
    #         {u'name': u'test-account-1', u'transaction': 'first_transaction'},
    #         {u'name': u'test-account-1', u'transaction': 'second_transaction'},
    #         {u'name': u'test-account-1', u'transaction': 'third_first_transaction'},
    #         {u'name': u'test-account-2', u'transaction': 'fourth_first_transaction'},
    #         {u'name': u'test-account-2', u'transaction': 'fifth_first_transaction'},
    #         {u'name': u'test-account-2', u'transaction': 'sixth_first_transaction'},
    #     ]
    #     assert len(result) == 6
    #     assert result == expected


if __name__ == '__main__':
    unittest.main()



import pytest

from fiddler import FiddlerApi


def test_client_v1_creation_fail():
    with pytest.raises(ValueError) as e:
        client_v1 = FiddlerApi('', '', '')


def test_client_v2_creation_fail():
    with pytest.raises(ValueError) as e:
        client_v2 = FiddlerApi('', '', '', version=2)



class ResourceNotFound(Exception):
    def __init__(self, message='Resource not found'):
        self.message = message
        super().__init__(self.message)


class JSONException(RuntimeError):
    def __init__(self, status, message, stacktrace, logs):
        self.status = status
        self.message = message
        self.stacktrace = stacktrace
        self.logs = logs

    def __str__(self):
        return (
            f'{self.status}: {self.message} \n\n '
            f'{self.stacktrace} \n Server Logs:\n {self.logs}'
        )


class FiddlerException(Exception):
    """An exception raised by client API in case of errors."""

    pass


class MalformedSchemaException(Exception):
    def __init__(self, message='Core Fiddler Object has Malformed Schema'):
        self.message = message
        super().__init__(self.message)


class CommandOptionsError(Exception):
    """Generic exception for insufficient args or configuration."""

    def __init__(self, *args, **kwargs):
        pass



import re


def _name_check(name: str, max_length: int):
    if not name:
        raise ValueError(f'Invalid name {name}')
    if len(name) > max_length:
        raise ValueError(f'Name longer than {max_length} characters: {name}')
    if len(name) < 2:
        raise ValueError(f'Name must be at least 2 characters long: {name}')
    if not bool(re.match('^[a-z0-9_]+$', name)):
        raise ValueError(
            f'Name must only contain lowercase letters, '
            f'numbers or underscore: {name}'
        )
    if name[0].isnumeric():
        raise ValueError(f'Name must not start with a numerical character: {name}')


def safe_name_check(name: str, max_length: int, strict_mode: bool = True):
    if strict_mode:
        _name_check(name, max_length)


def is_greater_than_max_value(value, limit, epsilon=1e-9):
    """Check if 'value' is significantly larger than 'limit'.
    Where significantly means theres a greater difference than epsilon.
    """
    max_arg = abs(max(value, limit))
    tolerance = max_arg * epsilon
    # make sure that a is "significantly" smaller than b before declaring it
    # identified in DI for hired
    return value - tolerance > limit


def is_less_than_min_value(value, limit, epsilon=1e-9):
    """Check if 'value' is significantly smaller than 'limit'.
    Where significantly means theres a greater difference than epsilon.
    """
    max_arg = abs(max(value, limit))
    tolerance = max_arg * epsilon
    # make sure that a is "significantly" smaller than b before declaring it
    # identified in DI for hired
    return value + tolerance < limit


def type_enforce(param_name: str, param_val, required_type):
    """
    :param param_name: Name of parameter being type enforced. Used for valid error
    :param param_val: Value being type enforced
    :param required_type: Casted type that is being enforced
    """
    if isinstance(param_val, required_type):
        return param_val  # should perhaps be required_type(param_val)?
    else:
        try:
            return required_type(param_val)
        except ValueError:
            raise ValueError(
                f'Parameter `{param_name}` must be of type {required_type}'
            )


def is_int_type(value):
    if isinstance(value, int):
        return True, value
    try:
        int_val = int(value)
        return True, int_val
    except ValueError:
        return False, None


def do_not_proceed(query: str):
    """
    Returns True if the users inputs n/no, False for yes/y

    :param query: Message displayed to the user

    Raises ValueError for invalid inputs
    """
    user_str = input(query)
    user_str = user_str.strip().lower()
    if user_str in ['y', 'yes']:
        return False
    elif user_str in ['n', 'no']:
        return True
    else:
        err_msg = 'Invalid response to the prompt, expecting one of y, n'
        raise ValueError(err_msg)



import re
import numpy as np
import pandas as pd
from fiddler import gem


class IGTextAttributionsTF2Keras:
    """
    Helper class for project attribution method when computing IG for Text data with Keras TF2.
    """
    def __init__(self, input_df, output_cols):
        """
        :param input_df: pandas dataframe for a single observation
        :param output_cols: list of output column names
        """
        self.input_df = input_df
        self.output_cols = output_cols

    def text_to_tokens_keras(self, tokenizer, max_seq_length, feature_label):
        """
        Helper function to convert text to tokens with keras TF2.

        :param tokenizer: keras tokenizer used during training
        :param max_seq_length: max sequence length used during training
        :param feature_label: str. Name of the text feature input
        :return: word tokens
        """
        from tensorflow.keras.preprocessing.sequence import pad_sequences

        unpadded_tokens = [
            tokenizer.texts_to_sequences([x])[0]
            for x in self.input_df[feature_label].values
        ]

        padded_tokens = pad_sequences(
            unpadded_tokens, max_seq_length, padding='post', truncating='post'
        )

        word_tokens = tokenizer.sequences_to_texts([[x] for x in padded_tokens[0]])
        return word_tokens

    def text_attributions(self, tokenizer, word_tokens, word_attributions, feature_label):
        """
        Helper function to define segments works and attributions.

        :param tokenizer: keras tokenizer used during training
        :param word_tokens: word tokens. Could be the output of the text_to_tokens_keras method
        :param word_attributions: associated word attributions
        :param feature_label: str. Name of the text feature input
        :return: final segments and attributions
        """
        segments = re.split(
            r'([ ' + tokenizer.filters + '])',
            self.input_df.iloc[0][feature_label],
        )
        i = 0
        final_attributions = []
        final_segments = []
        for segment in segments:
            if segment != '':  # dump empty tokens
                final_segments.append(segment)
                seg_low = segment.lower()
                if len(word_tokens) > i and seg_low == word_tokens[i]:
                    final_attributions.append(word_attributions[i])
                    i += 1
                else:
                    final_attributions.append(0)
        return final_segments, final_attributions

    def get_attribution_for_output(self, att, word_tokens, tokenizer, embedding_name, feature_label):
        """
        Helper function to get attributions for a given output.

        :param att: dictionary of attributions for the given output
        :param word_tokens: word tokens
        :param tokenizer: keras tokenizer used during training
        :param embedding_name: str. Name of the embedding layer in the model
        :param feature_label: str. Name of the text feature input
        :return:
        """
        # Note - summing over attributions in the embedding direction
        word_attributions = np.sum(att[embedding_name][-len(word_tokens):],
                                   axis=1)
        final_segments, final_attributions = self.text_attributions(tokenizer, word_tokens, word_attributions,
                                                                    feature_label)
        gem_text = gem.GEMText(feature_name=feature_label,
                               text_segments=final_segments,
                               text_attributions=final_attributions)

        return gem_text

    def get_project_attribution(self, attributions, tokenizer, word_tokens, embedding_name, feature_label):
        """
        Helper method to get project attributions when model has a single text input feature.

        :param attributions: list of IG attributions. Each element of the list corresponds to an output.
        :param tokenizer: tokenizer used during training
        :param word_tokens: word tokens
        :param embedding_name: str. Name of the embedding layer in the model
        :param feature_label: str. Name of the text feature input
        :return:
        """
        explanations_by_output = {}

        if isinstance(feature_label, list):
            if len(feature_label) == 1:
                feature_label = feature_label[0]
            else:
                raise ValueError("Your model has multiple inputs. You cannot use this helper. "
                                 "Please implement project_attributions accordingly. "
                                 "If you need some help, contact Fiddler.")
        if isinstance(embedding_name, list):
            if len(embedding_name) == 1:
                embedding_name = embedding_name[0]
            else:
                raise ValueError("Your model has multiple embeddings. You cannot use this helper. "
                                 "Please implement project_attributions accordingly. "
                                 "If you need some help, contact Fiddler.")

        for output_field_index, att in enumerate(attributions):
            gem_text = self.get_attribution_for_output(att, word_tokens, tokenizer, embedding_name, feature_label)
            gem_container = gem.GEMContainer(contents=[gem_text])
            explanations_by_output[self.output_cols[output_field_index]] = gem_container.render()

        return explanations_by_output


class IGTabularAttributions:
    """
    Helper class for project attribution method when computing IG for Tabular data.
    """

    def __init__(self, input_df, output_cols):
        """
        :param input_df: pandas dataframe for a single observation
        :param output_cols: list of output column names
        """
        self.input_df = input_df
        self.output_cols = output_cols

    def get_project_attribution(self, attributions, attr_input_names_mapping):
        """
        Helper method to get project attributions when input data is tabular.

        :param attributions: list of IG attributions. Each element of the list corresponds to an output.
        :param attr_input_names_mapping: dict that map attributable layer names to input feature names
        :return: dictionary explanations_by_output
        """
        explanations_by_output = {}
        if attr_input_names_mapping is None:
            raise ValueError('Parameter attr_input_names_mapping cannot be empty when IG is enabled.')
        for output_field_index, att in enumerate(attributions):
            explanations_by_output[self.output_cols[output_field_index]] = \
                get_tabular_attributions_for_output(self.input_df, att,
                                                    attr_input_names_mapping=attr_input_names_mapping)
        return explanations_by_output


class TreeShapAttributions:
    """
    Helper class for computing attributions with Tree SHAP for tree based model
    """

    def __init__(self, input_df, transformed_df, output_cols, model):
        """
        :param input_df: pandas dataframe for a single observation
        :param transformed_df: pandas dataframe for a single transformed observation
               (if any pre-processing steps are required before running the model on the data)
        :param output_cols: list of output column names
        """
        import shap

        self.explainer = shap.TreeExplainer(model)
        self.input_df = input_df
        self.transformed_df = transformed_df
        self.output_cols = output_cols

    def get_project_attribution(self, predictions_dict, baseline_pred_dict, transformed_names_mapping=None):
        """
        Helper method to get project attributions for Tree SHAP.

        :param predictions_dict: dictionary of predictions of the model for the given input
        :param baseline_pred_dict: mapping dictionary between output column names and corresponding
               baseline prediction value. The baseline value has to be chosen according to the problem.
               In some cases, it can be the mean prediction of the baseline dataset.
        :param transformed_names_mapping: dictionary mapping feature name to corresponding transformed columns
        :return: dictionaries explanations_by_output, extras_by_output
        """
        if not isinstance(baseline_pred_dict, dict):
            raise ValueError("The parameter baseline_pred_dict has to be a mapping dictionary between output col name "
                             "and baseline prediction value.")
        if set(baseline_pred_dict.keys()) != set(self.output_cols):
            raise ValueError("The parameter baseline_pred_dict has to have output col names as keys.")
        if not isinstance(predictions_dict, dict):
            raise ValueError("The parameter predictions_dict has to be a mapping dictionary between output col name "
                             "and prediction value for the given input.")
        if set(predictions_dict.keys()) != set(self.output_cols):
            raise ValueError("The parameter predictions_dict has to have output col names as keys.")

        shap_values = self.explainer.shap_values(self.transformed_df, check_additivity=False)
        explanations_by_output = {}
        extras_by_output = {}

        if (transformed_names_mapping is None) and (set(self.input_df.columns) != set(self.transformed_df.columns)):
            raise ValueError("Your data has some pre-processing steps before being fed to the model. "
                             "Please provide a dictionary with key feature names and values list of corresponding "
                             "transformed feature names.")

        for ind, output_name in enumerate(self.output_cols):
            if len(self.output_cols) == 1:
                if len(shap_values) == 2:
                    # For binary classification models that predict proba for 0 and 1
                    attributions = shap_values[1][0]
                else:
                    attributions = shap_values[0]
            else:
                attributions = shap_values[ind][0]
            if transformed_names_mapping is not None:
                # Map the transformed input names to shap values
                map_att_transformed = {name: attributions[ind] for ind, name in enumerate(self.transformed_df.columns)}
                # Map the original input names to sum of shap values
                attributions = {name: np.sum([map_att_transformed[col] for col in transformed_names_mapping[name]])
                                for name in transformed_names_mapping.keys()}
            explanations_by_output[output_name] = get_tabular_attributions_for_output(self.input_df,
                                                                                      attributions)

            extras_by_output[output_name] = {'model_prediction': predictions_dict[output_name],
                                             'baseline_prediction': baseline_pred_dict[output_name]}

        return explanations_by_output, extras_by_output


def get_tabular_attributions_for_output(input_df, attributions,
                                        attr_input_names_mapping=None):
    """
    Helper function to get attributions for a single output.

    :param input_df: pandas dataframe for a single observation
    :param attributions: dictionary of attributions with keys attributable input names if IG is enabled.
                         Otherwise, this is just a list of attributions.
    :param attr_input_names_mapping: Required only if IG is enabled.
    :return: explanations for a single output
    """
    dense_features = []

    if attr_input_names_mapping is None:
        if isinstance(attributions, (list, np.ndarray)):
            attributions = {'dense_inputs': attributions}
            attr_input_names_mapping = {'dense_inputs': list(input_df.columns)}
        elif isinstance(attributions, dict):
            attr_input_names_mapping = {'dense_inputs': list(attributions.keys())}
            attributions = {'dense_inputs': list(attributions.values())}
        else:
            raise ValueError(f"The argument attributions should be a list or a dictionary mapping."
                             f" It's currently of type: {type(attributions)}")

    for key in attributions.keys():
        if key not in attr_input_names_mapping:
            raise ValueError(f"The key {key} is missing in the attr_input_names_mapping dictionary.")
        for ind, feature_name in enumerate(attr_input_names_mapping[key]):
            attr_val = attributions[key][ind]
            if isinstance(attr_val, (list, np.ndarray)):
                attr_val = np.sum(attr_val)
            value = input_df[feature_name].iloc[0]
            if pd.isnull(value):
                value = None
            dense_features.append(
                gem.GEMSimple(
                    feature_name=feature_name,
                    value=value,
                    attribution=float(attr_val),
                )
            )
        # Attribute 0.0 for features not used in the model
        missing_features = set(input_df.columns) - set(attr_input_names_mapping[key])
        for feature_name in missing_features:
            value = input_df[feature_name].iloc[0]
            if pd.isnull(value):
                value = None
            dense_features.append(
                gem.GEMSimple(
                    feature_name=feature_name,
                    value=value,
                    attribution=float(0.0),
                )
            )
    gem_container = gem.GEMContainer(contents=dense_features)

    return gem_container.render()



import time
from http import HTTPStatus
from typing import List, Optional

from requests.exceptions import ConnectionError

from fiddler.libs.http_client import RequestClient
from fiddler.utils import logging
from fiddler.v2.schema.job import JobStatus
from fiddler.v2.utils.exceptions import handle_api_error_response
from fiddler.v2.utils.response_handler import JobResponseHandler

logger = logging.getLogger(__name__)


class JobMixin:
    client: RequestClient
    organization_name: str

    @handle_api_error_response
    def get_job(self, uuid: str) -> Optional[JobResponseHandler]:
        """
        Get details about a job

        :params uuid: Unique identifier of the job to get the details of
        :returns: JobResponseHandler object containing details
        """
        # Setting timeout as makeshift approach to handle pod restart
        # leading the ConnectionResetError on client
        timeout = 120
        response = self.client.get(url=f'jobs/{uuid}', timeout=timeout)
        if response.status_code == HTTPStatus.OK:
            return JobResponseHandler(response)
        else:
            # @TODO we should throw error here instead of logging
            logger.info(f'Response status code: {response.status_code}')
            logger.info(response.body)

    def poll_job(self, uuid: str, interval: int = 3) -> JobResponseHandler:
        """
        Poll an ongoing job. This method will keep polling until a job has SUCCESS, FAILURE, RETRY, REVOKED status.
        Will return a JobResponseHandler object once the job has ended.

        :params uuid: Unique identifier of the job to keep polling
        :params interval: Interval in sec between two subsequent poll calls. Default is 1sec
        :returns: JobResponseHandler object containing job details.
        """
        job = self.get_job(uuid)
        # @TODO: Set max iteration limit?
        while job.status in [JobStatus.PENDING, JobStatus.STARTED]:
            # @TODO: show proper message
            logger.info(
                f'JOB UUID: {uuid} status: {job.status} Progress: {job.progress:.2f}'
            )

            time.sleep(interval)
            try:
                job = self.get_job(uuid)
            except ConnectionError:
                logger.exception(
                    'Failed to connect to the server. Retrying...\nIf this error persists, please reach out to Fiddler.'
                )

        if job.status == JobStatus.FAILURE:
            logger.error(
                f"JOB UUID: {uuid} failed with error message '{job.error_message}'"
            )

        logger.info(
            f'JOB UUID: {uuid} status: {job.status} Progress: {job.progress:.2f}'
        )
        return job

    @staticmethod
    def get_task_results(job: JobResponseHandler) -> List[str]:
        results: List[str] = []
        if not job.extras:
            return results
        for task_id, extra_info in job.extras.items():
            if not extra_info['result']:
                continue
            results.append(
                f'JOB UUID: {job.uuid} task id: {task_id} result: {extra_info["result"]}'
            )
        return results



import math
import os
from typing import Dict, Optional
from urllib.parse import urljoin

from fiddler.libs.http_client import RequestClient
from fiddler.utils import logging
from fiddler.v2.constants import UploadType
from fiddler.v2.utils.response_handler import APIResponseHandler

logger = logging.getLogger(__name__)

UPLOAD_TYPE_KEY = 'upload_type'
RESOURCE_IDENT_KEY = 'resource_identifier'
FILE_NAME_KEY = 'file_name'
MULTIPART_UPLOAD_KEY = 'multipart_upload'
CONTENT_TYPE_KEY = 'Content-Type'
UPLOAD_ID_KEY = 'upload_id'


def multipart_upload(
    client: RequestClient,
    organization_name: str,
    project_name: str,
    identifier: str,
    upload_type: UploadType,
    file_path: str,
    file_name: str,
    chunk_size: Optional[int] = 100 * 1024 * 1024,  # bytes
) -> Dict[str, str]:

    base_url = f'datasets/{organization_name}:{project_name}/'
    UPLOAD_INITIALIZE_URL = urljoin(base_url, 'initialize')
    UPLOAD_URL = urljoin(base_url, 'upload')
    UPLOAD_COMPLETE_URL = urljoin(base_url, 'complete')

    if not os.path.exists(file_path):
        raise ValueError(
            f'File {file_path} not found. Please check if the entered path is correct.'
        )
    file_size = os.path.getsize(file_path)
    if file_size < (5 * 1024 * 1024):
        with open(file_path, 'rb') as f:
            file_data = f.read()
            response = client.put(
                url=UPLOAD_URL,
                params={
                    UPLOAD_TYPE_KEY: upload_type,
                    MULTIPART_UPLOAD_KEY: False,
                    RESOURCE_IDENT_KEY: identifier,
                    FILE_NAME_KEY: file_name,
                },
                data=file_data,
                headers={CONTENT_TYPE_KEY: 'application/octet-stream'},
            )
    else:
        # Step 1: Initialize multipart upload
        response = client.post(
            url=UPLOAD_INITIALIZE_URL,
            data={
                RESOURCE_IDENT_KEY: identifier,
                FILE_NAME_KEY: file_name,
                UPLOAD_TYPE_KEY: upload_type,
            },
        )
        response_data = APIResponseHandler(response).get_data()
        upload_id = response_data.get(UPLOAD_ID_KEY)
        file_name = response_data.get(FILE_NAME_KEY)
        part_no = 1
        parts = []
        total_chunks = math.ceil(file_size / chunk_size)
        # Step 2: Chunk and upload
        with open(file_path, 'rb') as f:
            logger.info('Beginning chunk upload')
            file_data = f.read(chunk_size)
            while file_data:
                response = client.put(
                    url=UPLOAD_URL,
                    params={
                        UPLOAD_TYPE_KEY: upload_type,
                        MULTIPART_UPLOAD_KEY: True,
                        RESOURCE_IDENT_KEY: identifier,
                        FILE_NAME_KEY: file_name,
                        UPLOAD_ID_KEY: upload_id,
                        'part_number': part_no,
                    },
                    data=file_data,
                    headers={CONTENT_TYPE_KEY: 'application/octet-stream'},
                )
                resp_data = APIResponseHandler(response).get_data()
                parts.append(
                    {
                        'ETag': resp_data.get('info').get('ETag'),
                        'PartNumber': resp_data.get('info').get('PartNumber'),
                    }
                )
                logger.info(f'Chunk uploaded: {part_no}/{total_chunks}')
                part_no += 1
                file_data = f.read(chunk_size)
        # Step 3: Complete the upload
        response = client.post(
            url=UPLOAD_COMPLETE_URL,
            params={
                UPLOAD_TYPE_KEY: upload_type,
                RESOURCE_IDENT_KEY: identifier,
                FILE_NAME_KEY: file_name,
                UPLOAD_ID_KEY: upload_id,
            },
            data={'parts': parts},
        )
    logger.info('File successfully uploaded to blob store')
    return APIResponseHandler(response).get_data()



import copy
import os
from typing import Dict, List, Optional

import pandas as pd

from fiddler.connection import Connection
from fiddler.core_objects import DatasetInfo
from fiddler.utils import df_from_json_rows
from fiddler.utils.general_checks import type_enforce


class Dataset:
    def __init__(self, connection: Connection, project_id: str, dataset_id: str):
        self.connection = connection
        self.project_id = project_id
        self.dataset_id = dataset_id

    def get_info(self) -> DatasetInfo:
        """Get DatasetInfo for a dataset.

        :returns: A fiddler.DatasetInfo object describing the dataset.
        """
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        dataset_id = type_enforce('dataset_id', self.dataset_id, str)

        path = ['dataset_schema', self.connection.org_id, project_id, dataset_id]
        res = self.connection.call(path, is_get_request=True)
        info = DatasetInfo.from_dict(res)
        info.dataset_id = dataset_id
        return info

    def _query_dataset(
        self,
        fields: List[str],
        max_rows: int,
        split: Optional[str] = None,
        sampling=False,
        sampling_seed=0.0,
    ):
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        dataset_id = type_enforce('dataset_id', self.dataset_id, str)

        payload = dict(
            fields=fields,
            limit=max_rows,
            sampling=sampling,
        )

        if sampling:
            payload['sampling_seed'] = sampling_seed
        if split is not None:
            payload['source'] = f'{split}.csv'

        path = [
            'dataset_query',
            self.connection.org_id,
            project_id,
            dataset_id,
        ]
        res = self.connection.call(path, json_payload=payload, stream=True)
        return res

    def download(
        self,
        max_rows: int = 1_000,
        splits: Optional[List[str]] = None,
        sampling=False,
        dataset_info: Optional[DatasetInfo] = None,
        include_fiddler_id=False,
    ) -> Dict[str, pd.DataFrame]:
        """Fetches data from a dataset on Fiddler.

        :param max_rows: Up to this many rows will be fetched from eash split
            of the dataset.
        :param splits: If specified, data will only be fetched for these
            splits. Otherwise, all splits will be fetched.
        :param sampling: If True, data will be sampled up to max_rows. If
            False, rows will be returned in order up to max_rows. The seed
            will be fixed for sampling.∂
        :param dataset_info: If provided, the API will skip looking up the
            DatasetInfo (a necessary precursor to requesting data).
        :param include_fiddler_id: Return the Fiddler engine internal id
            for each row. Useful only for debugging.

        :returns: A dictionary of str -> DataFrame that maps the name of
            dataset splits to the data in those splits. If len(splits) == 1,
            returns just that split as a dataframe, rather than a dataframe.
        """

        if dataset_info is None:
            dataset_info = self.get_info()
        else:
            dataset_info = copy.deepcopy(dataset_info)

        def get_df_from_split(split, fiddler_id=include_fiddler_id):
            column_names = dataset_info.get_column_names()
            if fiddler_id:
                column_names.insert(0, '__fiddler_id')
            dataset_rows = self._query_dataset(
                fields=column_names,
                max_rows=max_rows,
                split=split,
                sampling=sampling,
            )
            return df_from_json_rows(
                dataset_rows, dataset_info, include_fiddler_id=include_fiddler_id
            )

        if splits is None:
            use_splits = [
                os.path.splitext(filename)[0] for filename in dataset_info.files
            ]
        else:
            use_splits = splits
        res = {split: get_df_from_split(split) for split in use_splits}
        if splits is not None and len(splits) == 1:
            # unwrap single-slice results
            res = next(iter(res.values()))
        return res

    def delete(self):
        """Permanently delete a dataset.

        :returns: Server response for deletion action.
        """
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        dataset_id = type_enforce('dataset_id', self.dataset_id, str)

        path = ['dataset_delete', self.connection.org_id, project_id, dataset_id]
        result = self.connection.call(path)

        return result



from typing import List, Optional

from fiddler.connection import Connection
from fiddler.dataset import Dataset
from fiddler.model import Model
from fiddler.utils.general_checks import type_enforce


class Project:
    def __init__(self, connection: Connection, project_id: str):
        self.connection = connection
        self.project_id = project_id

    def list_models(self) -> List[str]:
        """List the names of all models in a project.

        :returns: List of strings containing the ids of each model in the
            specified project.
        """
        project_id = type_enforce('project_id', self.project_id, str)
        path = ['list_models', self.connection.org_id, project_id]
        res = self.connection.call(path, is_get_request=True)
        return res

    def delete(self):
        """Permanently delete a project.
        :returns: Server response for deletion action.
        """
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)
        path = ['delete_project', self.connection.org_id, project_id]
        result = self.connection.call(path)
        return result

    def share(
        self,
        role: str,
        user_name: Optional[str] = None,
        team_name: Optional[str] = None,
    ):
        """Share a project with other users and/or teams.

        :param role: one of ["READ", "WRITE", "OWNER"].
        :param user_name: (optional) username, typically an email address.
        :param team_name: (optional) name of the team.

        :returns: Server response for creation action.
        """
        if user_name is None and team_name is None:
            err = 'one of user_name, team_name must be provided'
            raise ValueError(err)

        if user_name is not None and team_name is not None:
            err = 'Only one of user_name or team_name must be provided'
            raise ValueError(err)

        if role not in ['READ', 'WRITE', 'OWNER']:
            err = 'role must be one of READ, WRITE, or OWNER'
            raise ValueError(err)

        payload = {
            'role': role,
            'user_name': user_name,
            'team_name': team_name,
        }

        path = ['apply_project_role', self.connection.org_id, self.project_id]
        return self.connection.call(path, json_payload=payload)

    def unshare(
        self,
        role: str,
        user_name: Optional[str] = None,
        team_name: Optional[str] = None,
    ):
        """un-Share a project with other users and/or teams.

        :param role: one of ["READ", "WRITE", "OWNER"].
        :param user_name: (optional) username, typically an email address.
        :param team_name: (optional) name of the team.

        :returns: Server response for creation action.
        """
        if user_name is None and team_name is None:
            err = 'one of user_name, team_name must be provided'
            raise ValueError(err)

        if user_name is not None and team_name is not None:
            err = 'Only one of user_name or team_name must be provided'
            raise ValueError(err)

        if role not in ['READ', 'WRITE', 'OWNER']:
            err = 'role must be one of READ, WRITE, or OWNER'
            raise ValueError(err)

        payload = {
            'role': role,
            'user_name': user_name,
            'team_name': team_name,
        }

        path = ['delete_project_role', self.connection.org_id, self.project_id]
        return self.connection.call(path, json_payload=payload)

    def list_roles(self):
        """List the users and teams with access to a given project.

        :returns: list of users and teams with access to a given project.
        """
        path = ['roles', self.connection.org_id, self.project_id]
        return self.connection.call(path, is_get_request=True)

    def list_datasets(self) -> List[str]:
        """List the ids of all datasets in the organization.

        :returns: List of strings containing the ids of each dataset.
        """
        # Type enforcement
        project_id = type_enforce('project_id', self.project_id, str)

        path = ['list_datasets', self.connection.org_id, project_id]
        res = self.connection.call(path, is_get_request=True)

        return res

    def model(self, model_id: str) -> Model:
        return Model(self.connection, self.project_id, model_id)

    def dataset(self, dataset_id: str) -> Dataset:
        return Dataset(self.connection, self.project_id, dataset_id)



from http import HTTPStatus
from typing import List, Optional

from pydantic import parse_obj_as

from fiddler.libs.http_client import RequestClient
from fiddler.utils import logging
from fiddler.v2.schema.model import Model
from fiddler.v2.utils.exceptions import handle_api_error_response
from fiddler.v2.utils.response_handler import (
    APIResponseHandler,
    PaginatedResponseHandler,
)

logger = logging.getLogger(__name__)


class ModelMixin:
    client: RequestClient
    organization_name: str

    @handle_api_error_response
    def get_models(self, project_name: str) -> List[Model]:
        """
        Get list of all models belonging to a project

        :params project_name: The project for which you want to get the models
        :returns: List containing Model objects
        """
        response = self.client.get(
            url='models',
            params={
                'organization_name': self.organization_name,
                'project_name': project_name,
            },
        )
        _, items = PaginatedResponseHandler(response).get_pagination_details_and_items()
        return parse_obj_as(List[Model], items)

    @handle_api_error_response
    def get_model(self, project_name: str, model_name: str) -> Model:
        """
        Get the details of a model.

        :params project_name: The project to which the model belongs to
        :params model_name: The model name of which you need the details
        :returns: Model object which contains the details
        """
        response = self.client.get(
            url=f'models/{self.organization_name}:{project_name}:{model_name}',
        )
        response_handler = APIResponseHandler(response)
        return Model.deserialize(response_handler)

    @handle_api_error_response
    def add_model(
        self,
        project_name: str,
        model_name: str,
        info: Optional[dict] = None,
        model_type: Optional[str] = None,
        framework: Optional[str] = None,
        requirements: Optional[str] = None,
    ) -> Model:
        """
        Function to add a model to fiddler for monitoring

        :param project_name: project name where the model will be added
        :type project_name: string
        :param model_name: name of the model
        :type model_name: string
        :param info: model related information passed as dictionary from user
        :type info: dictionary
        :param model_type: the type of model e.g. REGRESSION, BINARY CLASSIFICATION
        :type model_type: string
        :return: Model object which contains the model details
        """
        # @TODO: file_list required while setting up model?
        # @TODO: Handle model_deployment_id?
        request_body = Model(
            organization_name=self.organization_name,
            project_name=project_name,
            name=model_name,
            info=info,
            model_type=model_type,
            framework=framework,
            requirements=requirements,
        ).dict()
        response = self.client.post(
            url='models',
            data=request_body,
        )
        logger.info(f'{model_name} setup successful')

        return Model.deserialize(APIResponseHandler(response))

    @handle_api_error_response
    def delete_model(self, project_name: str, model_name: str) -> None:
        """
        Delete a model

        :params project_name: Project name to which the model belongs to.
        :params model_name: Model name to be deleted

        :returns: None
        """
        response = self.client.delete(
            url=f'models/{self.organization_name}:{project_name}:{model_name}',
        )
        if response.status_code == HTTPStatus.OK:
            logger.info(f'{model_name} delete request received.')
        else:
            # @TODO: Handle non 200 status response
            logger.info('Delete unsuccessful')



import logging
from http import HTTPStatus
from typing import Any, Dict, Optional

from fiddler.connection import Connection
from fiddler.project import Project

from ..core_objects import (
    InitMonitoringModifications,
    possible_init_monitoring_modifications,
)
from ..utils.general_checks import type_enforce

LOG = logging.getLogger()


class Monitoring:
    def __init__(self, connection: Connection, project_id: str, model_id: str) -> None:
        self.connection = connection
        self.project_id = project_id
        self.model_id = model_id

    def initialize_monitoring(  # noqa
        self,
        enable_modify: Optional[bool] = False,
        verbose: Optional[bool] = False,
        version: str = 'fiddler2',
    ):
        """
        Ensure that monitoring has been setup and Fiddler is ready to ingest events.

        :param enable_modify:       Grant the Fiddler backend permission to
                                    modify model related objects, schema, etc.

                                    Can be bool  `True`/`False`, indicating global
                                    write/read-only permission.

                                    Alternatively, can be a sequence of elements
                                    from `fiddler.core_objects.InitMonitoringModifications`,
                                    in which case all listed elements will be
                                    assumed to be modifiable, and omitted ones
                                    will be read-only.
        :param verbose:             Bool indicating whether to run in verbose
                                    mode with longer debug messages for errors.
        :param version:             String (either 'fiddler2' or 'fiddler1') indicating whether to use legacy functionality (with enable_modify and verbose) or new functionality (is_sync)

        :returns bool indicating whether monitoring could be setup correctly, or task_id if is_sync is True.
        """
        # TODO: only trigger model predictions if they do not already exist
        model_info = None
        dataset_id = None
        try:
            # model_info = self.get_model_info(self.project_id, self.model_id)
            model_info = (
                Project(self.connection, self.project_id)
                .model(self.model_id)
                .get_info()
            )
        except Exception as e:
            LOG.exception(
                f'Did not find ModelInfo for project "{self.project_id}" and model "{self.model_id}".'
            )
            raise e
        try:
            assert model_info is not None and model_info.datasets is not None  # nosec
            dataset_id = model_info.datasets[0]
            assert dataset_id is not None  # nosec
        except Exception as e:  # TODO: don't catch all exceptions.
            LOG.exception(
                f'Unable to infer dataset from model_info for given project_id={self.project_id} and model_id={self.model_id}. Did your model_info specify a dataset?'
            )
            if verbose:
                LOG.exception(
                    f'The inferred dataset_id was:\n{dataset_id}\n\nThe inferred model_info was:\n{model_info}'
                )
            raise e
        self._check_and_trigger_predictions(dataset_id)
        if version.lower() == 'fiddler2':
            path = [
                'init_monitoring',
                self.connection.org_id,
                self.project_id,
                self.model_id,
            ]
            payload = {}
            payload['version'] = 'fiddler2'
            """
            For Q1 2022, we will poll on the task-id returned from
            celery to check the status of the sketch generation task

            init_monitoring will thus appear synchronous
            """
            init_result = self.connection.call(path, json_payload=payload)
            if init_result['status'] != HTTPStatus.ACCEPTED:
                return False
            else:
                return init_result['task_id']
        else:
            pass
        default_modify = False

        if isinstance(enable_modify, bool):
            default_modify = enable_modify

        enable_backend_modifications = {
            check.value: default_modify
            for check in possible_init_monitoring_modifications
        }

        if isinstance(enable_modify, list) or isinstance(enable_modify, tuple):
            for mod in enable_modify:
                if type(mod) == InitMonitoringModifications:
                    mod = mod.value
                enable_backend_modifications[mod] = True

        elif type(enable_modify) == bool:
            pass
        else:
            raise NotImplementedError

        overall_result = True
        init_result = self._init_monitoring_api_call(
            verbose, enable_backend_modifications
        )

        overall_result = overall_result and init_result['success']
        # For now, we only permit precomputation if no failures are found # TODO: allow precomputations for incomplete schema?
        if init_result['success'] is False:
            return False
        if verbose:
            LOG.info('Precomputing Dataset Histograms')

        # TODO : This call is no longer needed, as initialize_monitoring is not supported for F1, and precompute call is only
        # relevant for F1
        precompute_result = self._precompute_api_call(verbose)
        overall_result = overall_result and precompute_result['success']

        if verbose:
            LOG.info(
                f'overall_result: {overall_result},\n\t- init_result: {init_result},\n\t- precompute_result: {precompute_result}'
            )
        return overall_result

    def _precompute_api_call(self, verbose: Optional[bool] = False):
        precompute_result = None
        try:
            path = [
                'precompute',
                self.connection.org_id,
                self.project_id,
                self.model_id,
            ]
            payload = {
                # 'dataset': dataset_id,
            }
            precompute_result = self.connection.call(path, json_payload=payload)
        except Exception as e:
            LOG.exception('Failed to precompute histograms, error message: ')
            raise e
        if precompute_result is not None:
            if precompute_result['success']:
                if verbose:
                    LOG.info('Successfully precomputed histograms, details: ')
                    LOG.info(precompute_result['message'])
            else:
                LOG.info('Failed to precompute histograms, details: ')
                LOG.info(precompute_result['message'])
        else:
            LOG.info(
                'Failed to precompute histograms, could not parse server response.'
            )
            precompute_result = {'success': False}
        return precompute_result

    def _init_monitoring_api_call(
        self,
        verbose: Optional[bool] = False,
        enable_backend_modifications: Optional[Dict] = None,
    ) -> dict:
        init_result = None
        try:
            path = [
                'init_monitoring',
                self.connection.org_id,
                self.project_id,
                self.model_id,
            ]
            payload = {}
            payload['enable_modifications'] = enable_backend_modifications

            """
            for fiddler2, set async to true in order to ensure that sketch
            generation happens in the celery worker

            For Q1 2022, we will poll on the task-id returned from
            celery to check the status of the sketch generation task
            """
            payload['version'] = 'fiddler1'
            init_result = self.connection.call(path, json_payload=payload)
        except Exception as e:
            LOG.exception('Failed to setup monitoring, error message: ')
            raise e
        if init_result is not None:
            if init_result['success']:
                if verbose:
                    LOG.warning(f"ERRORS: {init_result['errors']}")
                    LOG.warning(f"MESSAGE: {init_result['message']}")
            else:
                LOG.warning(f"ERRORS: {init_result['errors']}")
                LOG.warning(f"MESSAGE: {init_result['message']}")
        else:
            LOG.warning('Failed to setup monitoring, could not parse server response.')
            init_result = {'success': False}
        return init_result

    def _check_and_trigger_predictions(self, dataset_id):
        predictions_exist = False
        try:
            path = ['model_predictions_exist', self.connection.org_id, self.project_id]
            payload: Dict[str, Any] = {
                'model': self.model_id,
                'dataset': dataset_id,
            }  # is dataset_id same as dataset_name?
            predictions_exist = self.connection.call(path, json_payload=payload)
        except Exception as e:
            LOG.warning('Failed to check for predictions, regenerating by default')
            LOG.warning(str(e))
        if not predictions_exist:
            Project(self.connection, self.project_id).model(
                self.model_id
            )._trigger_model_predictions(dataset_id)
        else:
            LOG.info(
                f'Predictions already exist for model={self.model_id}, dataset={dataset_id}'
            )

    def add_monitoring_config(
        self,
        config_info: dict,
    ):
        """Adds a config for either an entire org, or project or a model.
        Here's a sample config:
        {
            'min_bin_value': 3600, # possible values 300, 3600, 7200, 43200, 86400, 604800 secs
            'time_ranges': ['Day', 'Week', 'Month', 'Quarter', 'Year'],
            'default_time_range': 7200,
            'tag': 'anything you want',
            ‘aggregation_config’: {
               ‘baseline’: {
                  ‘type’: ‘dataset’,
                  ‘dataset_name’: yyy
               }
            }
        }
        """
        # Type enforcement
        project_id = (
            type_enforce('project_id', self.project_id, str)
            if self.project_id
            else self.project_id
        )
        model_id = (
            type_enforce('model_id', self.model_id, str)
            if self.model_id
            else self.model_id
        )

        path = ['monitoring_setup', self.connection.org_id]
        if project_id:
            path.append(project_id)
        if model_id:
            if not project_id:
                raise ValueError(
                    'We need to have a `project_id` when a model is specified'
                )
            path.append(model_id)

        result = self.connection.call(path, config_info)
        self._dataset_baseline_display_message(result)
        return result

    def _dataset_baseline_display_message(self, config):
        try:
            aggregation_config_baseline = config['aggregation_config']['baseline']
            aggregation_config_baseline_type = aggregation_config_baseline['type']
            if aggregation_config_baseline_type == 'dataset':
                LOG.info(
                    'Dataset baseline will only be monitoring columns that are both present in the training data '
                    'and the specified dataset used as baseline'
                )
        except KeyError:
            pass



import re
import sys
from datetime import datetime

from .. import constants


def prettyprint_number(number, n_significant_figures=4):
    n_digits = len(f'{number:.0f}')
    return f'{round(number, n_significant_figures - n_digits):,}'


def _compute_hash(string):
    hash_value = 0
    if not string:
        return hash_value
    for c in string:
        hash_value = ((hash_value << 5) - hash_value) + ord(c)
        hash_value = hash_value & 0xFFFFFFFF  # Convert to 32bit integer
    return str(hash_value)


def sanitized_name(name):
    if name.isnumeric():
        name = f'_{name}'
    name = re.sub(r'[^a-zA-Z0-9_]', '_', name).lower()
    if len(name) > 63:
        suffix = f'_{_compute_hash(name)}'
        name = f'{name[0: 63 - len(suffix)]}{suffix}'
    return name


def validate_sanitized_names(columns, sanitized_name_dict):
    if not columns:
        return
    if isinstance(columns, str):
        columns = [columns]
    for column in columns:
        sname = sanitized_name(column.name)
        if sname in sanitized_name_dict:
            other_name = sanitized_name_dict.get(sname)
            raise ValueError(
                f'Name conflict, {column.name} and {other_name} '
                f'both maps to {sname}'
            )
        else:
            sanitized_name_dict[sname] = column.name


def pad_timestamp(str_ts) -> str:
    """
    Attempts to return a padded timestamp of format '%Y-%m-%d %H:%M:%S.%f'.
    Will pad with 0's as necessary.
    """
    # 2021-01-01 00:00:00.000000
    if re.match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}', str_ts):
        ts = str_ts
    # 2021-01-01 00:00:00.0+
    elif re.match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d+', str_ts):
        # Case of <6 floating point seconds
        ts = str_ts
        while not re.match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6}', ts):
            ts = f'{ts}0'
    # 2021-01-01 00:00:00
    elif re.match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}', str_ts):
        # Case of no floating point seconds
        ts = f'{str_ts}.000000'
    # 2021-01-01
    elif re.match(r'\d{4}-\d{2}-\d{2}', str_ts):
        ts = f'{str_ts} 00:00:00.000000'
    # If it doesn't match any other format, then we just mark as a failure
    else:
        ts = str_ts

    return ts


def formatted_utcnow(milliseconds=None) -> str:
    """:return: UTC timestamp in '%Y-%m-%d %H:%M:%S.%f' format."""
    if milliseconds:
        if type(milliseconds) != int:
            raise ValueError(
                f'Timestamp has to be provided in milliseconds '
                f'as an integer. Provided timestamp was '
                f'{milliseconds} of {type(milliseconds)}'
            )
        return datetime.utcfromtimestamp(milliseconds / 1000.0).strftime(
            constants.TIMESTAMP_FORMAT
        )
    return datetime.utcnow().strftime(constants.TIMESTAMP_FORMAT)


def print_streamed_result(res: str):
    if res.startswith(constants.ONE_LINE_PRINT):
        print(f'\r{res[len(constants.ONE_LINE_PRINT):]}', end='')
        sys.stdout.flush()
    else:
        print(res)


def model_string_to_hostname(name):
    # Allow only a-z, 0-9, and _
    return re.sub(r'[^a-zA-Z0-9]', '-', name).lower()



import pathlib
import re

import setuptools

# parse version number
version_regex = r"^__version__ = ['\"]([^'\"]*)['\"]"
version_file_contents = pathlib.Path('fiddler', '_version.py').open().read()
version_match = re.search(version_regex, version_file_contents, re.M)
if version_match:
    version = version_match.group(1)
else:
    raise RuntimeError('Unable to find version string.')


with open('PUBLIC.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='fiddler-client',
    version=version,
    author='Fiddler Labs',
    description='Python client for Fiddler Service',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://fiddler.ai',
    packages=setuptools.find_packages(),
    install_requires=[
        'pip>=21.0',
        'requests<3',
        'requests-toolbelt',
        'pandas>=1.2.5',
        'pyyaml',
        'packaging',
        'deepdiff',
        'boto3',
        'botocore',
        'fastavro',
        'importlib-resources',
        'Werkzeug',
        'pyarrow>=3.0.0',
        'deprecation==2.1.0',
        'pydantic==1.9.0',  # v2 dependency
        'deprecated==1.2.13',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    python_requires='>3.6.3',
    entry_points={
        'console_scripts': [
            'far = fiddler.archive.far:main',
        ],
    },
)



import numpy as np
import tensorflow as tf
from tensorflow.keras.models import Model


class ComputeGradientsKerasTF2:
    def __init__(self, model, model_dir, model_artifact_path, att_layer_names_mapping, embedding_names, hybrid=False):
        """
        Helper class to compute gradients for Keras TF2 models

        :param model: keras model
        :param model_dir: model directory pathlib path
        :param model_artifact_path: str. model artifact file name (if .h5 file format) or
               directory (if saved model format)
        :param att_layer_names_mapping: dict that map attributable layer names to input feature names.
               Values have to be list of feature name(s).
               For example: {'cat_var': ['country', 'gender'], 'num_var': ['price']}.
        :param embedding_names: List. Name(s) the embedding layer(s) of the model.
        :param hybrid: Boolean value for hybrid data type (for example matrices). Default to False.
        """
        self.model = model
        self.model_dir = model_dir
        self.model_artifact_path = model_artifact_path
        self.embedding_names = embedding_names

        if (att_layer_names_mapping is None) or (att_layer_names_mapping == {}):
            raise ValueError("att_layer_names_mapping argument cannot be empty."
                             " Please give the dictionary mapping of attributable layer names.")
        if not isinstance(att_layer_names_mapping, dict):
            raise ValueError("att_layer_names_mapping as to be a dictionary mapping between the attributable "
                             "layer names and the corresponding features names. "
                             "Values have to be list of feature name(s). "
                             "For example: {'cat_var': ['country', 'gender'], 'num_var': ['price']}")
        self.att_layer_names_mapping = att_layer_names_mapping
        if embedding_names is None:
            self.att_sub_models = None
        else:
            self.att_sub_models = self.get_attributable_sub_models()

        self.hybrid = hybrid

    def get_attributable_sub_models(self):
        """
        Construct sub-models for each attributable layer name. This method is used if there are embedding
        layers in the model because it's not possible to attribute directly to the input.

        :return: dictionary with keys the attributable layer names and values the sub models
        """
        return {att_layer: Model(self.model.inputs, outputs=self.model.get_layer(att_layer).output)
                for att_layer in self.att_layer_names_mapping.keys()}

    def transform_to_attributable_input(self, input_data):
        """
        This method is responsible for transforming the input dataframe to the upstream-most representation
        of model inputs that belongs to a continuous vector-space.
        For models with embedding layers (esp. NLP models) the first attributable layer is downstream of that.

        :param input_data: input data (pandas DataFrame when no embedding, otherwise outputs of the _transform_input)
        :return: dictionary with keys the attributable input layer names and values
               the differential input values associated.
        """
        if self.att_sub_models is None:
            if self.hybrid:
                return {att_layer_name: input_data[att_layer_name]
                        for att_layer_name, col in self.att_layer_names_mapping.items()}
            else:
                return {att_layer_name: input_data[col].values
                        for att_layer_name, col in self.att_layer_names_mapping.items()}
        else:
            return {att_layer: att_sub_model.predict(input_data) for att_layer, att_sub_model
                    in self.att_sub_models.items()}

    def define_model_grads(self, batch_input_shape_list):
        """
        Define a differentiable model, cut from the Embedding Layers.
        This will take as input what the transform_to_attributable_input function defined.

        :param batch_input_shape_list: List of tuples corresponding to the output shape of the embedding layer(s)
               of the model. For example if the model has two embeddings, the list should have two tuples.
        :return: Differentiable keras model
        """
        if batch_input_shape_list is None:
            raise ValueError("Please specify the batch_input_shape_list. It corresponds to the output shape"
                             " of the embedding layer(s).")
        if not isinstance(batch_input_shape_list, list):
            batch_input_shape_list = [batch_input_shape_list]

        model = tf.keras.models.load_model(
            str(self.model_dir / self.model_artifact_path)
        )

        for index, name in enumerate(self.embedding_names):
            model.layers.remove(model.get_layer(name))
            model.layers[index]._batch_input_shape = batch_input_shape_list[index]
            model.layers[index]._dtype = 'float32'
            model.layers[index]._name = name

        new_model = tf.keras.models.model_from_json(model.to_json())

        for layer in new_model.layers:
            try:
                layer.set_weights(self.model.get_layer(name=layer.name).get_weights())
            except:
                pass

        return new_model

    @staticmethod
    def gradients_input(x, grad_model):
        """
        Method to compute gradients.

        :param x: attributable input tensor
        :param grad_model: Differentiable keras model. If the model has embedding layer, please use define_model_grads
               to get the grad model. Otherwise, grad_model is simply the original model.
        :return: dictionary of gradients
        """
        with tf.GradientTape() as tape:
            tape.watch(x)
            preds = grad_model(x)

        grads = tape.gradient(preds, x)

        return grads

    def compute_gradients(self, attributable_input, grad_model):
        """
        This method computes gradients of the model output wrt to the differentiable input.
        In the backend, this method receives the output of the transform_to_attributable_input()
        method.

        :param attributable_input: dictionary with keys attributable layer names and values array of
               corresponding values. If there are embeddings, the attributable_input values should be the
               output of the embedding layer.
        :param grad_model: Differentiable keras model. If the model has embedding layer, please use define_model_grads
               to get the grad model. Otherwise, grad_model is simply the original model.
        :return: a list of dictionaries, where each entry of the list is the attribution
                 for an output.
                 In case of single output model, this is a list with a single entry.
                 For the dictionary, the keys are the name of the input layers and the values are the attributions.
        """
        gradients_by_output = []
        attributable_input_tensor = {
            k: tf.identity(v) for k, v in attributable_input.items()
        }
        gradients_dic_tf = self.gradients_input(attributable_input_tensor, grad_model)
        gradients_dic_numpy = dict(
            [key, np.asarray(value)] for key, value in gradients_dic_tf.items()
        )
        gradients_by_output.append(gradients_dic_numpy)
        return gradients_by_output
